/*
bASCII - an ASCII art generator GIMP plug-in
Copyright (C) 2005  Mate Kovacs

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Library General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#define PLUG_NAME "bASCII"
#define PROC_NAME "file_bascii_save"
#define PREVTEXT "<MMMM> <Ii|l> 0123456789ABCDEF"

/*
NOTE:
This is the programming style suggested for GIMP plugins, not mine.
So don't blame on me if you don't like it! :)
*/

static void query (void);

static void run (const gchar     *name,
                 gint            n_params,
                 const GimpParam *param,
                 gint            *nreturn_vals,
                 GimpParam       **return_vals);
								 
typedef struct
{
	gint w,h;
	guchar *data;
} map;

inline void get_graymap (map *dst, GimpDrawable *drawable);
inline void get_fontmap (map *dst, const char *fontname);
inline int  ascii_gen   (gchar *filename, map gm,map fm);

#include "ascii_gen.c"

GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,  /* init */
  NULL,  /* quit */
  query,
  run,
};

MAIN ()

static void
query (void)
{
  static GimpParamDef args[] =
  {
		{GIMP_PDB_INT32,    "run_mode",   "Interactive, non-interactive"},
		{GIMP_PDB_IMAGE,    "image",      "Image"},
		{GIMP_PDB_DRAWABLE, "drawable",   "Drawable to save"},
		{GIMP_PDB_STRING,   "filename",   "File to save in"},
		{GIMP_PDB_STRING,   "raw_name",   "Name entered"},
		{GIMP_PDB_STRING,   "font_name",  "Font name"}
  };

  gimp_install_procedure (
		PROC_NAME,
		"creates ASCII art for specified font",
		"you know, it's like... ASCII art :)",
		"Mate Kovacs",
		"Mate Kovacs <mkovaxx gmail com>",
		"2005",
		PLUG_NAME,
		"GRAY*",
		GIMP_PLUGIN, G_N_ELEMENTS (args), 0, args, NULL);

	gimp_register_save_handler (PROC_NAME, "txt,asc", "");
}

static void run(const gchar     *name,
                gint            n_params,
                const GimpParam *param,
                gint            *nreturn_vals,
                GimpParam       **return_vals)
{
	GimpRunMode       run_mode = param[0].data.d_int32;
	gint32            image_ID = param[1].data.d_int32;
	GimpDrawable      *drawable = gimp_drawable_get (param[2].data.d_drawable);
	gchar             *filename = param[3].data.d_string, *fontname = 0;
	
	GtkWidget *dlg;
	gboolean ok = TRUE;
	
	map fontmap = {0}, graymap = {0};
	
	GimpPDBStatusType status = GIMP_PDB_SUCCESS;
	static GimpParam values[1];
	
  *nreturn_vals = 1;
  *return_vals  = values;

	switch (run_mode)
	{
		case GIMP_RUN_NONINTERACTIVE:
			if(n_params != 6)
			{
				ok = FALSE;
				status = GIMP_PDB_CALLING_ERROR;
			}
			else fontname = param[5].data.d_string;
			break;

		case GIMP_RUN_INTERACTIVE:
			gimp_ui_init (PLUG_NAME, FALSE);
			dlg = gtk_font_selection_dialog_new (PLUG_NAME);
			gtk_font_selection_dialog_set_preview_text (
				(GtkFontSelectionDialog*)dlg, PREVTEXT);
			gtk_widget_show (dlg);
			ok = (gtk_dialog_run (GTK_DIALOG (dlg)) == GTK_RESPONSE_OK);
			fontname = gtk_font_selection_dialog_get_font_name (
				(GtkFontSelectionDialog*)dlg);
			gtk_widget_destroy (dlg);
			break;

		case GIMP_RUN_WITH_LAST_VALS:
			fontname = 0;
			break;
	}
	
	if(ok)
	{
		get_fontmap (&fontmap, fontname);
		get_graymap (&graymap, drawable);
		ascii_gen (filename, graymap,fontmap);
		free(fontmap.data);
		free(graymap.data);
	}

	gimp_drawable_detach (drawable);

  values[0].type = GIMP_PDB_STATUS;
  values[0].data.d_status = status;
}

inline void get_graymap (map *dst, GimpDrawable *drawable)
{
	gint         w = gimp_drawable_width (drawable->drawable_id),
               h = gimp_drawable_height (drawable->drawable_id);
	GimpPixelRgn pixreg;
	
	dst->data = (guchar*)malloc(w*h);
	
	gimp_pixel_rgn_init (&pixreg, drawable, 0,0, w,h, FALSE,FALSE);
	gimp_pixel_rgn_get_rect (&pixreg, dst->data, 0,0, w,h);
	
	dst->w = w;
	dst->h = h;
}

inline void get_fontmap (map *dst, const char *fontname)
{
	GdkPixmap            *pixmap = 0;
	int                  w = 0,h = 0, x = 0,y = 0;
	GdkGC                *gc = 0;
	PangoContext         *context = gdk_pango_context_get ();
	PangoLayout          *layout = pango_layout_new (context);
	PangoFontDescription *fontdesc = pango_font_description_from_string (fontname);
	GdkColor             black, white;
	GdkPixbuf            *pixbuf = 0;
	guchar               *srcbuf = 0;
	gint                 stride = 0;
	char                 valid_chars[95*2];
	
	for(x = 0; x<95; x++)
	{
		valid_chars[2*x  ] = x+32;
		valid_chars[2*x+1] = '\n';
	}
	
	pango_layout_set_font_description (layout, fontdesc);
	pango_layout_set_text (layout, valid_chars, 95*2-1);
	pango_layout_get_pixel_size (layout, &w, &h);
	pixmap = gdk_pixmap_new (NULL,w,h,24);
	gc = gdk_gc_new (pixmap);
	black.pixel = 0;
	white.pixel = 0xFFFFFFFF;
	/*
	black.red = black.green = black.blue = 0x0000;
  white.red = white.green = white.blue = 0xFFFF;
	*/	
	gdk_gc_set_foreground (gc,&white);
	gdk_draw_rectangle (pixmap,gc,TRUE, 0,0, w,h);
  gdk_gc_set_foreground (gc,&black);
	gdk_gc_set_background (gc,&white);
	gdk_draw_layout (pixmap,gc, 0,0, layout);
	pango_font_description_free (fontdesc);
	
	pixbuf = gdk_pixbuf_get_from_drawable (NULL,pixmap,NULL, 0,0, 0,0, -1,-1);
	srcbuf = gdk_pixbuf_get_pixels (pixbuf);
	stride = gdk_pixbuf_get_rowstride(pixbuf);
	
	dst->data = (guchar*)malloc(w*h);
	
	for(y=0; y<h; y++)
		for(x=0; x<w; x++)
			dst->data[x + y*w] = srcbuf[x*3 + y*stride];
	
	dst->w = w;
	dst->h = h/95;
}
