/*
 * gimp-image-reg -- A GIMP Plugin for Image Registration
 *
 * dialogs.c
 * Copyright (C) 2010 Behnam Tabatabai
 *
 * This file is part of gimp-image-reg.
 *
 * gimp-image-reg is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * gimp-image-reg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with gimp-image-reg. If not, see <http://www.gnu.org/licenses/>.
 */

#include "dialog.h"

#include "config.h"

#include "defs.h"
#include "i18n.h"
#include "render.h"
#include "user-data.h"

#include <libgimp/gimpui.h>
#include <string.h>

#define RESPONSE_RESET  1
#define RESPONSE_CANCEL 2
#define RESPONSE_OK     3

/* ------------------------------------------------------------------
 * Data types.
 * ------------------------------------------------------------------ */
typedef struct {
    GtkWidget* ref_drawable;
    GtkWidget* matching_area[2];
    GtkWidget* transformation[2];
    GtkWidget* interpolation[4];
    GtkWidget* clip_method[4];
    GtkObject* num_iterations;
    GtkObject* accuracy;
    GtkWidget* reset_button;
    GtkWidget* cancel_button;
    GtkWidget* ok_button;
} WidgetSet;

typedef struct {
    gint32    this_drawable_id;
    UserData  user_data;
    WidgetSet widget_set;
    gboolean  dialog_result;
} DialogData;

/* ------------------------------------------------------------------
 * Prototypes of local functions.
 * ------------------------------------------------------------------ */
static GtkWidget* CreateDialog(DialogData*);

static void RefreshDialog(DialogData*);

static void Callback_Drawable(GtkWidget* widget, gpointer data);

static void Callback_Response(
    GtkWidget*  widget,
    gint        response_id,
    DialogData* dialog_data);

static gboolean FilterDrawable(
    gint32   image_id,
    gint32   drawable_id,
    gpointer data);

/* ------------------------------------------------------------------
 * Callback functions.
 * ------------------------------------------------------------------ */
static void Callback_Drawable(GtkWidget* widget, gpointer data)
{
    gboolean success;
    gint value;
    DialogData* dialog_data;

    dialog_data = (DialogData*) data;
    
    success = gimp_int_combo_box_get_active(
        (GimpIntComboBox*) widget,
        &value);
    
    if (success)
    {
        dialog_data->user_data.ref_drawable = value;
    }

    gtk_widget_set_sensitive(
        dialog_data->widget_set.ok_button,
        gimp_drawable_is_valid(dialog_data->user_data.ref_drawable));
}

static void Callback_Response(
    GtkWidget*  widget,
    gint        response_id,
    DialogData* dialog_data)
{
    switch (response_id)
    {
        case RESPONSE_RESET:
        {
            UserData_GetDefaults(&dialog_data->user_data);
            RefreshDialog(dialog_data);
            
            dialog_data->dialog_result = FALSE;
        } break;

        case RESPONSE_OK:
        {
            gtk_widget_destroy(GTK_WIDGET(widget));
            dialog_data->dialog_result = TRUE;
        } break;
      
        default:
        {
            gtk_widget_destroy(GTK_WIDGET(widget));
            dialog_data->dialog_result = FALSE;
        } break;
    }
}

/* ------------------------------------------------------------------
 * Constraint functions.
 * ------------------------------------------------------------------ */

static gboolean FilterDrawable(
    gint32 image_id,
    gint32 drawable_id,
    gpointer data)
{
    DialogData* dialog_data;
    dialog_data = (DialogData*) data;
    return drawable_id != dialog_data->this_drawable_id;
}

/* ------------------------------------------------------------------
 * Update functions.
 * ------------------------------------------------------------------ */
static void RefreshDialog(DialogData* dialog_data)
{
    /* Radio buttons. */
    {
        gint32 image_id;
        gint x1, y1, x2, y2;
        gint x, y, w, h;
        gboolean non_empty_1;
        gboolean non_empty_2;
        gboolean success;

        image_id = gimp_drawable_get_image(dialog_data->this_drawable_id);
        success = gimp_selection_bounds(
            image_id,
            &non_empty_1,
            &x1, &y1,
            &x2, &y2);
        non_empty_2 = gimp_drawable_mask_intersect(
            dialog_data->this_drawable_id,
            &x, &y, &w, &h);
        if (success && non_empty_1 && non_empty_2)
        {
            /* Set to whatever value we have got. */
            gtk_toggle_button_set_active(
                (GtkToggleButton *)(dialog_data->widget_set.matching_area[
                                        dialog_data->user_data.matching_area]),
                TRUE);
        }
        else
        {
            /* Select the option "whole image". */
            gtk_toggle_button_set_active(
                (GtkToggleButton *)
                (dialog_data->widget_set.matching_area[
                    MATCHING_AREA_WHOLE_IMAGE]),
                TRUE);

            /* Disable the option "selection". */
            gtk_widget_set_sensitive(
                dialog_data->widget_set.matching_area[
                    MATCHING_AREA_SELECTION],
                FALSE);
        }
    }

    
    gtk_toggle_button_set_active(
        (GtkToggleButton *)(dialog_data->widget_set.transformation[
                                dialog_data->user_data.transformation]),
        TRUE);

    gtk_toggle_button_set_active(
        (GtkToggleButton *)(dialog_data->widget_set.interpolation[
                                dialog_data->user_data.interpolation]),
        TRUE);

    gtk_toggle_button_set_active(
        (GtkToggleButton *)(dialog_data->widget_set.clip_method[
                                dialog_data->user_data.clip_method]),
        TRUE);

    /* Num iterations and accuracy. */
    {
        gtk_adjustment_set_value(
            GTK_ADJUSTMENT(dialog_data->widget_set.num_iterations),
            dialog_data->user_data.num_iterations);
        
        gtk_adjustment_set_value(
            GTK_ADJUSTMENT(dialog_data->widget_set.accuracy),
            dialog_data->user_data.accuracy);
        
    }

    /* Reference image. */
    {
        gboolean success;
        gint     d;

        if (gimp_drawable_is_valid(dialog_data->user_data.ref_drawable))
        {
            gimp_int_combo_box_set_active(
                (GimpIntComboBox*)(dialog_data->widget_set.ref_drawable),
                dialog_data->user_data.ref_drawable);
        }
                
        success = gimp_int_combo_box_get_active(
            (GimpIntComboBox*)(dialog_data->widget_set.ref_drawable),
            &d);

        if (success)
        {
            dialog_data->user_data.ref_drawable = d;
        }

        gtk_widget_set_sensitive(
            dialog_data->widget_set.ok_button,
            gimp_drawable_is_valid(dialog_data->user_data.ref_drawable));
    
    }
}

/* ------------------------------------------------------------------
 * Create and run the dialog.
 * ------------------------------------------------------------------ */
gboolean RunDialog(GimpDrawable* drawable, UserData* user_data)
{
    DialogData dialog_data;
    GtkWidget* dialog;

    /* Initialize everything in the dialog data to zero. */
    memset(&dialog_data, 0, sizeof(DialogData));

    /* Make a copy of user data (as defaults). */
    dialog_data.user_data = *user_data;

    /* Initialize the drawable id. */
    dialog_data.this_drawable_id = drawable->drawable_id;

    gimp_ui_init(PLUGIN_BINARY, 0);

    dialog = CreateDialog(&dialog_data);
    if (!dialog)
    {
        return FALSE;
    }

    gtk_widget_show(dialog);

    gtk_main();

    if (dialog_data.dialog_result)
    {
        *user_data = dialog_data.user_data;
    }
    return dialog_data.dialog_result;
}

GtkWidget* CreateDialog(DialogData* dialog_data)
{
    GtkWidget* dialog;
    GtkWidget* main_box;

    dialog = gimp_dialog_new(
        _("Image Registration"),
        PLUGIN_BINARY,
        NULL,
        0,
        gimp_standard_help_func,
        PLUGIN_NAME,
        /* Buttons will be added by explicit calls to get access to
         * corresponding widgets. See below.*/
        NULL);

    if (!dialog)
    {
        return NULL;
    }

    /* Add buttons. */
    dialog_data->widget_set.reset_button = gimp_dialog_add_button(
        GIMP_DIALOG(dialog),
        GIMP_STOCK_RESET,
        RESPONSE_RESET);

    dialog_data->widget_set.cancel_button = gimp_dialog_add_button(
        GIMP_DIALOG(dialog),
        GTK_STOCK_CANCEL,
        RESPONSE_CANCEL);

    dialog_data->widget_set.ok_button = gimp_dialog_add_button(
        GIMP_DIALOG(dialog),
        GTK_STOCK_OK,
        RESPONSE_OK);

    gtk_dialog_set_alternative_button_order(
        GTK_DIALOG(dialog),
        RESPONSE_RESET,
        RESPONSE_OK,
        RESPONSE_CANCEL,
        -1);

    gimp_window_set_transient(GTK_WINDOW(dialog));
    main_box = gtk_vbox_new(FALSE, 12);
    gtk_container_set_border_width(GTK_CONTAINER(main_box), 12);
    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), main_box);

    /* -----------------------------------------------------------
     * Reference Image
     * ----------------------------------------------------------- */
    {
        GtkWidget* frame;
        GtkWidget* combo;
        frame = gimp_frame_new(_("Reference Image"));
        gtk_box_pack_start(GTK_BOX (main_box), frame, FALSE, FALSE, 0);

        combo = gimp_layer_combo_box_new(FilterDrawable, dialog_data);
        dialog_data->widget_set.ref_drawable = combo;
        g_signal_connect(
            combo,
            "changed",
            G_CALLBACK(Callback_Drawable),
            dialog_data);

        gtk_container_add(GTK_CONTAINER(frame), combo);
        gtk_widget_show(combo);
        gtk_widget_show(frame);

        gtk_widget_set_tooltip_text(
            frame,
            _("Reference image"));
        
        gtk_widget_set_tooltip_text(
            combo,
            _("Reference image"));
    }

    /* -----------------------------------------------------------
     * Matching algorithm.
     * ----------------------------------------------------------- */
    {
        GtkWidget* table;
        GtkWidget* frame;
         
        table = gtk_table_new(1, 2, FALSE);
        gtk_table_set_col_spacings(GTK_TABLE(table), 0);
        gtk_table_set_row_spacings(GTK_TABLE(table), 0);
        gtk_box_pack_start(GTK_BOX (main_box), table, FALSE, FALSE, 0);
        gtk_widget_show(table);


        /* Transformation. */
        frame = gimp_int_radio_group_new(
            TRUE,
            _("Transformation Model"),
            G_CALLBACK (gimp_radio_button_update),
            &dialog_data->user_data.transformation,
            dialog_data->user_data.transformation,

            _("Shift and Rotate"),
            XFORMATION_RIGID,
            &dialog_data->widget_set.transformation[XFORMATION_RIGID],

            _("Affine Transformation"),
            XFORMATION_AFFINE,
            &dialog_data->widget_set.transformation[XFORMATION_AFFINE],

            NULL);
        gtk_widget_show(frame);
        gtk_table_attach_defaults(
            GTK_TABLE(table),
            frame,
            0, 1,
            0, 1);

        /* Transformation tooltips. */
        gtk_widget_set_tooltip_text(
            frame,
            _("How to transform the image"));
        
        gtk_widget_set_tooltip_text(
            dialog_data->widget_set.transformation[XFORMATION_RIGID],
            _("Translation and rotation (3 degrees of freedom)"));
        
        gtk_widget_set_tooltip_text(
            dialog_data->widget_set.transformation[XFORMATION_AFFINE],
            _("Affine transformation (6 degrees of freedom)"));
        
        /* Matching area. */
        frame = gimp_int_radio_group_new(
            TRUE,
            _("Matching Area"),
            G_CALLBACK (gimp_radio_button_update),
            &dialog_data->user_data.matching_area,
            dialog_data->user_data.matching_area,

            _("Whole Image"),
            MATCHING_AREA_WHOLE_IMAGE,
            &dialog_data->widget_set.matching_area[MATCHING_AREA_WHOLE_IMAGE],

            _("Selection"),
            MATCHING_AREA_SELECTION,
            &dialog_data->widget_set.matching_area[MATCHING_AREA_SELECTION],

            NULL);
        gtk_widget_show(frame);
        gtk_table_attach_defaults(
            GTK_TABLE(table),
            frame,
            1, 2,
            0, 1);

        /* Matching area tooltips. */
        gtk_widget_set_tooltip_text(
            frame,
            _("Region used as matching area"));
        
        gtk_widget_set_tooltip_text(
            dialog_data->widget_set.matching_area[MATCHING_AREA_WHOLE_IMAGE],
            _("The whole image"));
        
        gtk_widget_set_tooltip_text(
            dialog_data->widget_set.matching_area[MATCHING_AREA_SELECTION],
            _("Only the selected area"));
        
    }

    /* -----------------------------------------------------------
     * Calculation of matching parameters.
     * ----------------------------------------------------------- */
    {
        const gint SCALE_WIDTH       = 128;
        const gint SPIN_BUTTON_WIDTH = 64;
        
        gint row = 0;
        GtkWidget* table;
        GtkObject* adj;

        /* Create a table for all scaler objects. */
        table = gtk_table_new(3, 3, FALSE);
        gtk_table_set_col_spacings(GTK_TABLE(table), 6);
        gtk_table_set_row_spacings(GTK_TABLE(table), 2);
        gtk_box_pack_start(GTK_BOX (main_box), table, FALSE, FALSE, 0);
        gtk_widget_show(table);

        /* Minimum Required Accuracy. */
        adj = gimp_scale_entry_new(
            GTK_TABLE (table), 0, row++,
            _("Accuracy 10^(-n)"),
            SCALE_WIDTH, SPIN_BUTTON_WIDTH,
            dialog_data->user_data.accuracy, /* initial value */
            3,                          /* lower boundary */
            12,                         /* upper boundary */
            1,                          /* step increment */
            4,                          /* page increment */
            0,                          /* digits */
            TRUE,                       /* constrain */                     
            0,                          /* unconstrained lower */
            0,                          /* unconstrained upper */
            _("Stop as soon the error is less than 10^(-n)"), /* tooltip */
            NULL                        /* help-id */ );
        dialog_data->widget_set.accuracy = adj;
        g_signal_connect(
            adj,
            "value_changed",
            G_CALLBACK(gimp_int_adjustment_update),
            &dialog_data->user_data.accuracy);
        
        /* Maximum Number of Iterations. */
        adj = gimp_scale_entry_new(
            GTK_TABLE(table), 0, row++,
            _("Iterations"),
            SCALE_WIDTH, SPIN_BUTTON_WIDTH,
            dialog_data->user_data.num_iterations, /* initial value */
            3,                          /* lower boundary */
            4096,                       /* upper boundary */
            1,                          /* step increment */
            16,                         /* page increment */
            0,                          /* digits */
            FALSE,                      /* constrain */                     
            1,                          /* unconstrained lower */
            0x7fffffff,                 /* unconstrained upper */
            _("Stop after so many iterations"), /* tooltip */
            NULL                        /* help-id */ );
        dialog_data->widget_set.num_iterations = adj;
        g_signal_connect(
            adj,
            "value_changed",
            G_CALLBACK(gimp_int_adjustment_update),
            &dialog_data->user_data.num_iterations);
    }
    
    /* -----------------------------------------------------------
     * How to apply the results of image alignment.
     * ----------------------------------------------------------- */
    {
        GtkWidget* frame;
        GtkWidget* vbox;
        GtkWidget* table;

        vbox = main_box;

        table = gtk_table_new(1, 2, FALSE);
        gtk_table_set_col_spacings(GTK_TABLE(table), 0);
        gtk_table_set_row_spacings(GTK_TABLE(table), 0);
        gtk_box_pack_start(GTK_BOX (vbox), table, FALSE, FALSE, 0);
        gtk_widget_show(table);


        /* Interpolation. */
        frame = gimp_int_radio_group_new(
            TRUE,
            _("Interpolation"),
            G_CALLBACK (gimp_radio_button_update),
            &dialog_data->user_data.interpolation,
            dialog_data->user_data.interpolation,

            _("None"),
            GIMP_INTERPOLATION_NONE,
            &dialog_data->widget_set.interpolation[GIMP_INTERPOLATION_NONE],

            _("Linear"),
            GIMP_INTERPOLATION_LINEAR,
            &dialog_data->widget_set.interpolation[GIMP_INTERPOLATION_LINEAR],

            _("Cubic"),
            GIMP_INTERPOLATION_CUBIC,
            &dialog_data->widget_set.interpolation[GIMP_INTERPOLATION_CUBIC],

            _("Sinc (Lanczos)"),
            GIMP_INTERPOLATION_LANCZOS,
            &dialog_data->widget_set.interpolation[GIMP_INTERPOLATION_LANCZOS],

            NULL);
        gtk_widget_show(frame);
        gtk_table_attach_defaults(
            GTK_TABLE(table),
            frame,
            0, 1,
            1, 2);


        /* Tooltips for interpolation. */
        gtk_widget_set_tooltip_text(
            frame,
            _("How to interpolate the transformed image"));

        gtk_widget_set_tooltip_text(
            dialog_data->widget_set.interpolation[GIMP_INTERPOLATION_NONE],
            _("No interpolation\n"
              "(take the nearest neighbor)"));
        
        gtk_widget_set_tooltip_text(
            dialog_data->widget_set.interpolation[GIMP_INTERPOLATION_LINEAR],
            _("Bi-linear interpolation"));
        
        gtk_widget_set_tooltip_text(
            dialog_data->widget_set.interpolation[GIMP_INTERPOLATION_CUBIC],
            _("Bi-cubic interpolation"));
        
        gtk_widget_set_tooltip_text(
            dialog_data->widget_set.interpolation[GIMP_INTERPOLATION_LANCZOS],
            _("Lanczos interpolation"));
        
        /* Clip method. */
        frame = gimp_int_radio_group_new(
            TRUE,
            _("Clipping"),
            G_CALLBACK (gimp_radio_button_update),
            &dialog_data->user_data.clip_method,
            dialog_data->user_data.clip_method,

            _("Adjust"),
            GIMP_TRANSFORM_RESIZE_ADJUST,
            &dialog_data->widget_set.clip_method[GIMP_TRANSFORM_RESIZE_ADJUST],

            _("Clip"),
            GIMP_TRANSFORM_RESIZE_CLIP,
            &dialog_data->widget_set.clip_method[GIMP_TRANSFORM_RESIZE_CLIP],

            _("Crop"),
            GIMP_TRANSFORM_RESIZE_CROP,
            &dialog_data->widget_set.clip_method[GIMP_TRANSFORM_RESIZE_CROP],

            _("Crop with Aspect"),
            GIMP_TRANSFORM_RESIZE_CROP_WITH_ASPECT,
            &dialog_data->widget_set.clip_method[
                GIMP_TRANSFORM_RESIZE_CROP_WITH_ASPECT],

            NULL);
        gtk_widget_show(frame);
        gtk_table_attach_defaults(
            GTK_TABLE(table),
            frame,
            1, 2,
            1, 2);

        /* Tooltips for clipping method. */
        gtk_widget_set_tooltip_text(
            frame,
            _("How to clip the transformed image"));
            
        gtk_widget_set_tooltip_text(
            dialog_data->widget_set.clip_method[GIMP_TRANSFORM_RESIZE_ADJUST],
            _("Enlarge the layer to contain the\n"
              "transformed image"));
        
        gtk_widget_set_tooltip_text(
            dialog_data->widget_set.clip_method[GIMP_TRANSFORM_RESIZE_CLIP],
            _("Delete all parts of the transformed image\n"
              "which exceed the image boundaries"));
        
        gtk_widget_set_tooltip_text(
            dialog_data->widget_set.clip_method[GIMP_TRANSFORM_RESIZE_CROP],
            _("Crop the layer to contain only the non-empty\n"
              "parts of the transformed image"));
        
        gtk_widget_set_tooltip_text(
            dialog_data->widget_set.clip_method[
                GIMP_TRANSFORM_RESIZE_CROP_WITH_ASPECT],
            _("Same as \"Crop\", but conserving the aspect\n"
              "ratio of the original image"));
    }

    RefreshDialog(dialog_data);

    gtk_widget_show(main_box);

    g_signal_connect(
        dialog,
        "response",
        G_CALLBACK(Callback_Response),
        dialog_data);
    
    g_signal_connect(
        dialog,
        "destroy",
        G_CALLBACK(gtk_main_quit),
        NULL);

    return dialog;

}

