/*
 * gimp-image-reg -- A GIMP Plugin for Image Registration
 *
 * image-ops.c
 * Copyright (C) 2010 Behnam Tabatabai
 *
 * This file is part of gimp-image-reg.
 *
 * gimp-image-reg is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * gimp-image-reg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with gimp-image-reg. If not, see <http://www.gnu.org/licenses/>.
 */

#include "image-ops.h"
#include <stdio.h>
#include <math.h>

#define USE_NEIGHBORS 0
#define USE_SCHARR    1

#define _SCHARR_X(k)                            \
    ( ( 3 * (*k[2] + *k[3]) + 10 * *k[6] ) -    \
      ( 3 * (*k[0] + *k[1]) + 10 * *k[4] ) )

#define _SCHARR_Y(k)                            \
    ( ( 3 * (*k[1] + *k[2]) + 10 * (*k[5]) ) -  \
      ( 3 * (*k[0] + *k[3]) + 10 * (*k[7]) ) )

#define _SOBEL_X(k)                             \
    ( ( *k[2] + *k[3] + 2 * *k[6] ) -           \
      ( *k[0] + *k[1] + 2 * *k[4] ) )
    
#define _SOBEL_Y(k)                             \
    ( ( *k[1] + *k[2] + 2 * *k[5] ) -           \
      ( *k[0] + *k[3] + 2 * *k[7] ) )

#if USE_SCHARR

# define DERIV_X(k)                                     \
    (gint16) floor(_SCHARR_X(kernel) / 32.0 + 0.5)

# define DERIV_Y(k) \
    (gint16) floor(_SCHARR_Y(kernel) / 32.0 + 0.5)

#else

# define DERIV_X(k)                                     \
    (gint16) floor(_SOBEL_X(kernel) / 8.0 + 0.5)

# define DERIV_Y(k) \
    (gint16) floor(_SOBEL_Y(kernel) / 8.0 + 0.5)

#endif

/* ------------------------------------------------------------------
 * ConvertToGrayscale
 * ------------------------------------------------------------------ */
static void ConvertToGrayscale(
    const guint8* pixel_data,
    gint32 count,
    gint32 bpp,
    guint8* dst)
{
    guint8* d;
    const guint8* r;
    const guint8* g;
    const guint8* b;

    const guint8* last_r;
    gint32 grayscale;

    last_r = pixel_data + count * bpp;

    for (d = dst,
         r = pixel_data,
         g = pixel_data + 1,
         b = pixel_data + 2;

         r < last_r;

         d += 1,
         r += bpp,
         g += bpp,
         b += bpp)
    {
        grayscale = (30 * (gint32)*r +
                     59 * (gint32)*g +
                     11 * (gint32)*b) / 100;
        *d = MIN(255, grayscale);
    }

}
/* ------------------------------------------------------------------
 * ReleaseImage
 * ------------------------------------------------------------------ */
void ReleaseImage(Image* image)
{
    if (image)
    {
        if (image->pixel)
        {
            DeleteMatrix_Uint8(image->pixel);
            image->pixel = NULL;
        }
        if (image->mask)
        {
            DeleteMatrix_Uint8(image->mask);
            image->mask = NULL;
        }
        if (image->gradient)
        {
            DeleteMatrix_Int16(image->gradient);
            image->gradient = NULL;
        }
        if (image->steepest_descent)
        {
            DeleteMatrix_Real(image->steepest_descent);
            image->steepest_descent = NULL;
        }
        if (image->hessian)
        {
            DeleteMatrix_Real(image->hessian);
            image->hessian = NULL;
        }
    }
}

/* ------------------------------------------------------------------
 * Extract_GrayScaleImage
 * ------------------------------------------------------------------ */
Matrix_Uint8* Extract_GrayScaleImage(
    GimpDrawable* drawable,
    gint32 x1,
    gint32 y1,
    gint32 width,
    gint32 height)
{
    gint32 drawable_id;
    gint32 x2;
    gint32 y2;

    GimpPixelRgn pixel_region;
    Matrix_Uint8* matrix;

    int x;
    int j;

    drawable_id = drawable->drawable_id;

    if (x1 < 0) return NULL;
    if (y1 < 0) return NULL;
    if (width < 1) return NULL;
    if (height < 1) return NULL;
    if (gimp_drawable_is_indexed(drawable_id)) return NULL;

    x2 = MIN(x1+width, drawable->width);
    y2 = MIN(y1+height, drawable->height);

    matrix = CreateMatrix_Uint8(
        y2 - y1,
        x2 - x1,
        1);

    if (!matrix) return NULL;


    gimp_pixel_rgn_init(
        &pixel_region,
        drawable,
        x1, y1,
        x2 - x1,
        y2 - y1,
        FALSE, FALSE);

    if (drawable->bpp == 1)
    {
        for (x = x1, j = 0; x < x2; ++x, ++j)
        {
            gimp_pixel_rgn_get_col(
                &pixel_region,
                ELEM(matrix, 0, j, 0),
                x, y1,
                y2 - y1);
        }
    }
    else
    {
        guint8* pixel_data;
        pixel_data = g_malloc((y2-y1) * sizeof(guint8) * drawable->bpp);

        if (!pixel_data)
        {
            DeleteMatrix_Uint8(matrix);
            return NULL;
        }

        for (x = x1, j = 0; x < x2; ++x, ++j)
        {
            gimp_pixel_rgn_get_col(
                &pixel_region,
                pixel_data,
                x, y1,
                y2 - y1);
            ConvertToGrayscale(
                pixel_data,
                y2 - y1,
                drawable->bpp,
                ELEM(matrix, 0, j, 0));
        }
        g_free(pixel_data);
    }

    return matrix;
}

/* ------------------------------------------------------------------
 * CreateGradient
 * ------------------------------------------------------------------ */
Matrix_Int16* CreateGradient(const Matrix_Uint8* image)
{
    /* Numbering scheme:
     *
     * 0   7   3
     * 4       6
     * 1   5   2
     *
     */

    typedef const guint8* const_pointer_uint8;
    const_pointer_uint8 kernel[8];
    const_pointer_uint8 k0;
    gint32 offsets[8];

#if (!USE_NEIGHBORS)
    static const guint8 s_outside = 255;
    const_pointer_uint8 outside = &s_outside;
#endif
    
    Matrix_Int16* gradient;
    gint16* gradient_x;
    gint16* gradient_y;

    gint i;
    gint r, c;

    if (NUM_ROWS(image) < 3) return NULL;
    if (NUM_COLS(image) < 3) return NULL;
    if (NUM_CHNS(image) != 1) return NULL;

    gradient = CreateMatrix_Int16(
        NUM_ROWS(image),
        NUM_COLS(image),
        2);

    if (!gradient) return NULL;

    offsets[0] = - NUM_ROWS(image) - 1;
    offsets[1] = - NUM_ROWS(image) + 1;
    offsets[2] =   NUM_ROWS(image) + 1;
    offsets[3] =   NUM_ROWS(image) - 1;
    offsets[4] = - NUM_ROWS(image);
    offsets[5] =   1;
    offsets[6] =   NUM_ROWS(image);
    offsets[7] = - 1;

    /* ------------------------
     * Upper left.
     * ------------------------ */
    k0 = ELEM(image, 0, 0, 0);
    for (i = 0; i < 8; ++i)
    {
        kernel[i] = k0 + offsets[i];
    }
#if USE_NEIGHBORS
    kernel[0] = k0;
    kernel[1] = kernel[5];
    kernel[3] = kernel[6];
    kernel[4] = k0;
    kernel[7] = k0;
#else
    kernel[0] = outside;
    kernel[1] = outside;
    kernel[3] = outside;
    kernel[4] = outside;
    kernel[7] = outside;
#endif
    gradient_x = ELEM(gradient, 0, 0, 0);
    gradient_y = gradient_x + 1;

    *gradient_x = DERIV_X(kernel);
    *gradient_y = DERIV_Y(kernel);


    /* ------------------------
     * Lower left.
     * ------------------------ */
    k0 = ELEM(image, NUM_ROWS(image)-1, 0, 0);
    for (i = 0; i < 8; ++i)
    {
        kernel[i] = k0 + offsets[i];
    }
#if USE_NEIGHBORS
    kernel[0] = kernel[7];
    kernel[1] = k0;
    kernel[2] = kernel[6];
    kernel[4] = k0;
    kernel[5] = k0;
#else
    kernel[0] = outside;
    kernel[1] = outside;
    kernel[2] = outside;
    kernel[4] = outside;
    kernel[5] = outside;
#endif    

    gradient_x = ELEM(gradient, NUM_ROWS(image)-1, 0, 0);
    gradient_y = gradient_x + 1;

    *gradient_x = DERIV_X(kernel);
    *gradient_y = DERIV_Y(kernel);
    
    /* ------------------------
     * Lower right.
     * ------------------------ */
    k0 = ELEM(image, NUM_ROWS(image)-1, NUM_COLS(image)-1, 0);
    for (i = 0; i < 8; ++i)
    {
        kernel[i] = k0 + offsets[i];
    }
#if USE_NEIGHBORS
    kernel[1] = kernel[4];
    kernel[2] = k0;
    kernel[3] = kernel[7];
    kernel[5] = k0;
    kernel[6] = k0;
#else
    kernel[1] = outside;
    kernel[2] = outside;
    kernel[3] = outside;
    kernel[5] = outside;
    kernel[6] = outside;
#endif
    gradient_x = ELEM(gradient, NUM_ROWS(image)-1, NUM_COLS(image)-1, 0);
    gradient_y = gradient_x + 1;

    *gradient_x = DERIV_X(kernel);
    *gradient_y = DERIV_Y(kernel);

    /* ------------------------
     * Upper right.
     * ------------------------ */
    k0 = ELEM(image, 0, NUM_COLS(image)-1, 0);
    for (i = 0; i < 8; ++i)
    {
        kernel[i] = k0 + offsets[i];
    }
#if USE_NEIGHBORS
    kernel[0] = kernel[4];
    kernel[2] = kernel[5];
    kernel[3] = k0;
    kernel[6] = k0;
    kernel[7] = k0;
#else
    kernel[0] = outside;
    kernel[2] = outside;
    kernel[3] = outside;
    kernel[6] = outside;
    kernel[7] = outside;
#endif
    gradient_x = ELEM(gradient, 0, NUM_COLS(image)-1, 0);
    gradient_y = gradient_x + 1;

    *gradient_x = DERIV_X(kernel);
    *gradient_y = DERIV_Y(kernel);
    
    /* ------------------------
     * Left column.
     * ------------------------ */
    for (r = 1; r < NUM_ROWS(image)-1; ++r)
    {
        k0 = ELEM(image, r, 0, 0);
        for (i = 0; i < 8; ++i)
        {
            kernel[i] = k0 + offsets[i];
        }
#if USE_NEIGHBORS
        kernel[0] = kernel[7];
        kernel[1] = kernel[5];
        kernel[4] = k0;
#else
        kernel[0] = outside;
        kernel[1] = outside;
        kernel[4] = outside;
#endif

        gradient_x = ELEM(gradient, r, 0, 0);
        gradient_y = gradient_x + 1;

        *gradient_x = DERIV_X(kernel);
        *gradient_y = DERIV_Y(kernel);
    }

    /* ------------------------
     * Right column.
     * ------------------------ */
    for (r = 1; r < NUM_ROWS(image)-1; ++r)
    {
        k0 = ELEM(image, r, NUM_COLS(image)-1, 0);
        for (i = 0; i < 8; ++i)
        {
            kernel[i] = k0 + offsets[i];
        }
#if USE_NEIGHBORS
        kernel[2] = kernel[5];
        kernel[3] = kernel[7];
        kernel[6] = k0;
#else
        kernel[2] = outside;
        kernel[3] = outside;
        kernel[6] = outside;
#endif        

        gradient_x = ELEM(gradient, r, NUM_COLS(image)-1, 0);
        gradient_y = gradient_x + 1;

        *gradient_x = DERIV_X(kernel);
        *gradient_y = DERIV_Y(kernel);
    }

    /* ------------------------
     * Upper row.
     * ------------------------ */
    for (c = 1; c < NUM_COLS(image)-1; ++c)
    {
        k0 = ELEM(image, 0, c, 0);
        for (i = 0; i < 8; ++i)
        {
            kernel[i] = k0 + offsets[i];
        }
#if USE_NEIGHBORS
        kernel[0] = kernel[4];
        kernel[3] = kernel[6];
        kernel[7] = k0;
#else
        kernel[0] = outside;
        kernel[3] = outside;
        kernel[7] = outside;
#endif

        gradient_x = ELEM(gradient, 0, c, 0);
        gradient_y = gradient_x + 1;

        *gradient_x = DERIV_X(kernel);
        *gradient_y = DERIV_Y(kernel);
    }

    /* ------------------------
     * Lower row.
     * ------------------------ */
    for (c = 1; c < NUM_COLS(image)-1; ++c)
    {
        k0 = ELEM(image, NUM_ROWS(image)-1, c, 0);
        for (i = 0; i < 8; ++i)
        {
            kernel[i] = k0 + offsets[i];
        }
#if USE_NEIGHBORS
        kernel[1] = kernel[4];
        kernel[2] = kernel[6];
        kernel[5] = k0;
#else
        kernel[1] = outside;
        kernel[2] = outside;
        kernel[5] = outside;
#endif

        gradient_x = ELEM(gradient, NUM_ROWS(image)-1, c, 0);
        gradient_y = gradient_x + 1;

        *gradient_x = DERIV_X(kernel);
        *gradient_y = DERIV_Y(kernel);
    }

    /* ------------------------
     * Inside image.
     * ------------------------ */
    for (c = 1; c < NUM_COLS(image)-1; ++c)
    {
        k0 = ELEM(image, 1, c, 0);
        for (i = 0; i < 8; ++i)
        {
            kernel[i] = k0 + offsets[i];
        }

        gradient_x = ELEM(gradient, 1, c, 0);
        gradient_y = gradient_x + 1;

        for (r = 1; r < NUM_ROWS(image)-1; ++r)
        {
            *gradient_x = DERIV_X(kernel);
            *gradient_y = DERIV_Y(kernel);
            
            /* Increment for the next row. */
            for (i = 0; i < 8; ++i)
            {
                kernel[i] += 1;
            }
            gradient_x += 2;
            gradient_y += 2;
        }
    }

    return gradient;
}

/* ------------------------------------------------------------------
 * Image_WritePgm
 * ------------------------------------------------------------------ */
gboolean Image_WritePgm(
    const gchar* name,
    const Matrix_Uint8* image)
{
    FILE* f;
    const guint8* pixel;
    const guint8* last_pixel;
    gint32 row;
    gint32 count;

    f = fopen(name, "w");
    if (!f) return FALSE;

    fprintf(f, "P2\n");
    fprintf(f, "# Created by Image Alignment Plugin\n");
    fprintf(f, "%d %d\n", NUM_COLS(image), NUM_ROWS(image));
    fprintf(f, "255\n");

    count = 0;
    for (row = 0; row < NUM_ROWS(image); ++row)
    {
        pixel      = ELEM(image, row, 0, 0);
        last_pixel = ELEM(image, row, NUM_COLS(image), 0);
        for (; pixel < last_pixel; pixel += NUM_ROWS(image))
        {
            ++count;
            fprintf(f, "%d", (gint32)(*pixel));
            if (count == 10)
            {
                count = 0;
                fprintf(f, "\n");
            }
            else
            {
                fprintf(f, " ");
            }
        }
    }

    fclose(f);
    return TRUE;
}

/* ------------------------------------------------------------------
 * Gradient_WritePgm
 * ------------------------------------------------------------------ */
gboolean Gradient_WritePgm(
    const gchar* name,
    const Matrix_Int16* image,
    gint channel)
{
    FILE* f;
    const gint16* pixel;
    const gint16* last_pixel;
    gint32 row;
    gint32 count;

    f = fopen(name, "w");
    if (!f) return FALSE;

    fprintf(f, "P2\n");
    fprintf(f, "# Created by Image Alignment Plugin\n");
    fprintf(f, "%d %d\n", NUM_COLS(image), NUM_ROWS(image));
    fprintf(f, "255\n");

    count = 0;
    for (row = 0; row < NUM_ROWS(image); ++row)
    {
        pixel      = ELEM(image, row, 0, channel);
        last_pixel = ELEM(image, row, NUM_COLS(image), channel);
        for (; pixel < last_pixel; pixel += NUM_ROWS(image) * 2)
        {
            ++count;
            fprintf(f, "%d", (gint32)(*pixel + 128));
            if (count == 10)
            {
                count = 0;
                fprintf(f, "\n");
            }
            else
            {
                fprintf(f, " ");
            }
        }
    }

    fclose(f);
    return TRUE;
}
