/*
 * gimp-image-reg -- A GIMP Plugin for Image Registration
 *
 * linalg.c
 * Copyright (C) 2010 Behnam Tabatabai
 *
 * This file is part of gimp-image-reg.
 *
 * gimp-image-reg is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * gimp-image-reg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with gimp-image-reg. If not, see <http://www.gnu.org/licenses/>.
 */

/*
c     ------------------------------------------------------------------
c     Source:
c     Numerical Recipes in Fortran 77: The Art of Scientific Computing
c     Second edition.
c     ------------------------------------------------------------------

c     ------------------------------------------------------------------
c     Given a matrix a(1:n, 1:n), with first physical dimension np, this
c     routine replaces it by the LU decomposition of a rowwise
c     permutation of itself. a and n are input. a is output, arranged as
c     in equation (2.3.14) above; indx(1:n) is an output vector that
c     records the row permutation effected by the partial pivoting; d is
c     output as +/-1 depending on whether the number of row interchanges
c     was even or odd, respectively. This routine is used in combination
c     with lubksb to solve linear equations of invert a matrix.
c     ------------------------------------------------------------------
*/

#include "linalg.h"

#include <math.h>
#include <stdlib.h>
#include <string.h>

#define LOCAL_BUFFER_SIZE 1024

gboolean LU_Decompose(
    Matrix_Real* a,
    gint* indx,
    gdouble* signum_out)
{
    gdouble  vv_local[LOCAL_BUFFER_SIZE];
    gdouble* vv;
    
    gdouble signum_local;
    gdouble* signum;

    static const gdouble tiny = 1e-20;

    gint i;
    gint j;
    gint k;
    gint imax;
    gint n;
    
    gdouble aamax;
    gdouble dum;
    gdouble sum;

    n = NUM_ROWS(a);
    
    /* For sizes less than LOCAL_BUFFER_SIZE use the local buffer. */
    if (n > LOCAL_BUFFER_SIZE)
    {
        vv = g_malloc(n * sizeof(gdouble));
        if (!vv)
        {
            return FALSE;
        }
    }
    else
    {
        vv = vv_local;
    }

    /* Keep the code easy by using a local signum if none is given. */
    if (signum_out)
    {
        signum = signum_out;
    }
    else
    {
        signum = &signum_local;
    }

    /* For the formal correctness. */
    imax = -1;

    /* No row interchanges yet. */
    *signum = 1.0;

    /* Loop over rows to get the implicit scaling information. */
    for (i = 0; i < n; ++i)
    {
        aamax = 0.0;
        for (j = 0; j < n; ++j)
        {
            if (fabs(*ELEM(a, i, j, 0)) > aamax)
            {
                aamax = fabs(*ELEM(a, i, j, 0));
            }
        }

        if (aamax == 0.0)
        {
            return FALSE;
        }
        vv[i] = 1.0 / aamax;
    }

    /* This is the loop over columns of Crout's method. */
    for (j = 0; j < n; ++j)
    {
        /* This is equation (2.3.12) except for i = j. */
        for (i = 0; i < j; ++i)
        {
            sum = *ELEM(a, i, j, 0);
            for (k = 0; k < i; ++k)
            {
                sum -= *ELEM(a, i, k, 0) * *ELEM(a, k, j, 0);
            }
            *ELEM(a, i, j, 0) = sum;
        }
        
        /* Initialize for the search for largest pivot element. */
        aamax = 0.0;

        /* This is i = j of equation (2.3.12) and i = j+1...N of
           equation (2.3.13). */

        for (i = j; i < n; ++i)
        {
            sum = *ELEM(a, i, j, 0);
            for (k = 0; k < j; ++k)
            {
                sum -= *ELEM(a, i, k, 0) * *ELEM(a, k, j, 0);
            }
            *ELEM(a, i, j, 0) = sum;

            /* Figure of merit for the pivot. */
            dum = vv[i] * fabs(sum);

            /* Is it better than the best so far? */
            if (dum >= aamax)
            {
                imax = i;
                aamax = dum;
            }
        }

        /* Do we need to interchange rows? */
        if (j != imax)
        {
            /* Yes, do so... */
            for (k = 0; k < n; ++k)
            {
                dum                  = *ELEM(a, imax, k, 0);
                *ELEM(a, imax, k, 0) = *ELEM(a, j,    k, 0);
                *ELEM(a, j,    k, 0) = dum;
            }

            /* ...and change the parity of signum. */
            *signum = - *signum;

            /* Also interchange the scale factor. */
            vv[imax] = vv[j];
        }

        indx[j] = imax;

        if (*ELEM(a, j, j, 0) == 0.0)
        {
            /* If the pivot element is zero the matrix is singular (at
               least to the precision of the algorithm). For some
               applications on singular matrices, it is desirable to
               substitute tiny for zero. */
            *ELEM(a, j, j, 0) = tiny;
        }

        /* Now, finally, divide by the pivot element. */
        if (j != n-1)
        {
            dum = 1.0 / *ELEM(a, j, j, 0);
            for (i = j+1; i < n; ++i)
            {
                *ELEM(a, i, j, 0) *= dum;
            }
        }
        
        /* Go back to the next column in the reduction. */
    }

    /* Release the buffer if allocated from heap. */
    if (vv != vv_local)
    {
        g_free(vv);
    }
    
    return TRUE;
}

/*
c     ------------------------------------------------------------------
c     Solves the set of n linear equations A . X = B. Here a is input,
c     not as the matrix A but rather as its LU decomposition, determined
c     by the routine ludcmp. indx is input as the permutation vector
c     returned by ludcmp. b(1:n) is input as the right-hand side vector
c     B, and returns with the solution vector X. a, n, np, and indx are
c     not modified by this routine and can be left in place for
c     successive calls with different right-hand sides b. This routine
c     take into account the possibility that b will begin with many zero
c     elements, so it is efficient for use in matrix inversion.
c     ------------------------------------------------------------------
*/
void LU_BackSubstitute(
    const Matrix_Real* a,
    const gint* indx,
    const gdouble* b,
    gdouble* x)
{
    gint i;
    gint ii;
    gint j;
    gint ll;
    gint n;
    gdouble sum;
    
    /* When ii is set to a non-negative value, if will become the
       index of the first nonvanishing element of b. We now do the
       forward substitution, equation (2.3.6). The only new wrinkle is
       to unscramble the permutation as we go. */

    n = NUM_ROWS(a);
    memcpy(x, b, n*sizeof(gdouble));
    
    ii = -1;
    for (i = 0; i < n; ++i)
    {
        ll = indx[i];
        sum = x[ll];
        x[ll] = x[i];
        if (ii >= 0)
        {
            for (j = ii; j < i; ++j)
            {
                sum -= *ELEM(a, i, j, 0) * x[j];
            }
        }
        else if (sum != 0.0)
        {
            /* A nonzero element was encountered, so from now on we
               will have to do the sums in the loop above. */
            ii = 0;
        }
        x[i] = sum;
    }

    /* Now we do the backsubstitution, equation (2.3.7). */
    for (i = n-1; i >= 0; --i)
    {
        sum = x[i];
        for (j = i+1; j < n; ++j)
        {
            sum -= *ELEM(a, i, j, 0) * x[j];
        }
        /* Store a component of the solution vector X. */
        x[i] = sum / *ELEM(a, i, i, 0);
    }

    /* All done! */
}
