/*
 * gimp-image-reg -- A GIMP Plugin for Image Registration
 *
 * main.c
 * Copyright (C) 2010 Behnam Tabatabai
 *
 * This file is part of gimp-image-reg.
 *
 * gimp-image-reg is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * gimp-image-reg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with gimp-image-reg. If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"

#include "dialog.h"
#include "defs.h"
#include "i18n.h"
#include "render.h"
#include "user-data.h"

#include <libgimp/gimp.h>
#include <stdio.h>

/* ------------------------------------------------------------------ */
/* Some utility functions */
/* ------------------------------------------------------------------ */
#if RELATIVE_DIRS
static char* Util_GetPersonalGimpDir(void)
{
    gchar* progdir;
    gchar* gimpdir;
    progdir = g_path_get_dirname(gimp_get_progname());
    if (!progdir)
    {
        return NULL;
    }
    
    gimpdir = g_path_get_dirname(progdir);
    g_free(progdir);

    return gimpdir;
}
#endif

static char* Util_GetDataDir(void)
{
#if RELATIVE_DIRS

    gchar* gimpdir;
    gchar* datadir;
        
    gimpdir = Util_GetPersonalGimpDir();
    if (!gimpdir)
    {
        return g_strdup(DATADIR);
    }
    
    datadir = g_build_filename(
        gimpdir,
        RELATIVE_DIR_SHARE,
        PLUGIN_NAME,
        NULL);
    g_free(gimpdir);
    
    if (!datadir)
    {
        return g_strdup(DATADIR);
    }
    
    return datadir;
#else
    return g_strdup(DATADIR);
#endif
}


static char* Util_GetLocaleDir(void)
{
#if RELATIVE_DIRS

    gchar* gimpdir;
    gchar* localedir;
        
    gimpdir = Util_GetPersonalGimpDir();
    if (!gimpdir)
    {
        return g_strdup(LOCALEDIR);
    }
    
    localedir = g_build_filename(
        gimpdir,
        RELATIVE_DIR_SHARE,
        RELATIVE_DIR_LOCALE,
        NULL);
    g_free(gimpdir);

    if (!localedir)
    {
        return g_strdup(LOCALEDIR);
    }
    
    return localedir;
#else
    return g_strdup(LOCALEDIR);
#endif
}

static void Util_SetGimpStatus(
    GimpParam* param,
    GimpPDBStatusType status)
{
    param->type = GIMP_PDB_STATUS;
    param->data.d_status = status;
}

static gboolean Util_CheckParameters(
    gint32 drawable_id,
    const UserData* user_data)
{
    if (!gimp_drawable_is_valid(drawable_id))
    {
        return FALSE;
    }

    if (!gimp_drawable_is_valid(user_data->ref_drawable))
    {
        return FALSE;
    }

    if (drawable_id == user_data->ref_drawable)
    {
        return FALSE;
    }

    if (user_data->matching_area < MATCHING_AREA_WHOLE_IMAGE ||
        user_data->matching_area > MATCHING_AREA_SELECTION)
    {
        return FALSE;
    }

    if (user_data->transformation < XFORMATION_RIGID ||
        user_data->transformation > XFORMATION_AFFINE)
    {
        return FALSE;
    }

    if (user_data->num_iterations < 1)
    {
        return FALSE;
    }

    if (user_data->accuracy < 3)
    {
        return FALSE;
    }

    if (user_data->interpolation < GIMP_INTERPOLATION_NONE ||
        user_data->interpolation > GIMP_INTERPOLATION_LANCZOS)
    {
        g_assert(!"Program error.");
        return FALSE;
    }

    if (user_data->clip_method < GIMP_TRANSFORM_RESIZE_ADJUST ||
        user_data->clip_method > GIMP_TRANSFORM_RESIZE_CROP_WITH_ASPECT)
    {
        g_assert(!"Program error.");
        return FALSE;
    }

    return TRUE;
}

static gboolean ReportRenderProgress(
    void* client_data,
    const gchar* message,
    gint32 iteration,
    gdouble accuracy,
    gdouble percentual_accuracy)
{
    gchar buffer[1024];

    gdouble* current_progress;
    current_progress = (gdouble*) client_data;

    if (iteration)
    {
        if (*current_progress < percentual_accuracy)
        {
            *current_progress = percentual_accuracy;
        }
    
        snprintf(
            buffer, sizeof(buffer),
            _("%s: [step %d, delta %7.3e]"),
            PACKAGE_NAME,
            iteration,
            accuracy);
    }
    else
    {
        *current_progress += 0.01;
        
        if (message)
        {
            snprintf(
                buffer, sizeof(buffer),
                _("%s: [%s]"),
                PACKAGE_NAME, message);
        }
        else
        {
            snprintf(
                buffer, sizeof(buffer),
                _("%s..."),
                PACKAGE_NAME);
        }
    }
    gimp_progress_set_text(buffer);
    gimp_progress_update(*current_progress);

    return TRUE;
}


/* ------------------------------------------------------------------
   Query and Run
   ------------------------------------------------------------------ */

static void query(void);
static void run(
    const gchar*     name,
    gint             nparams,
    const GimpParam* param,
    gint*            nreturn_vals,
    GimpParam**      return_vals);


GimpPlugInInfo PLUG_IN_INFO = {
    NULL,
    NULL,
    query,
    run
};

MAIN()

void query(void)
{
    static GimpParamDef args[] =
    {
        /* Parameter 0 */
        { GIMP_PDB_INT32,
          "run-mode",
          "Interactive, non-interactive." },

        /* Parameter 1 */
        { GIMP_PDB_IMAGE,
          "image",
          "Input image (not used)." },

        /* Parameter 2 */
        { GIMP_PDB_DRAWABLE,
          "drawable",
          "Input drawable." },

        /* Parameter 3 */
        { GIMP_PDB_DRAWABLE,
          "reference-drawable",
          "Reference drawable." },

        /* Parameter 4 */
        { GIMP_PDB_INT32,
          "matching-area",
          "What to use for image alignment:\n"
          "0 ... the whole image\n"
          "1 ... only the selected areas"
        },
        
        /* Parameter 5 */
        { GIMP_PDB_INT32,
          "transformation-model",
          "Transformation model used for aligning the images:\n"
          "0 ... move and rotate\n"
          "1 ... affine transformation"
        },
        
        /* Parameter 6 */
        { GIMP_PDB_INT32,
          "num-iterations",
          "Maximum number of iterations."},

        /* Parameter 7 */
        { GIMP_PDB_INT32,
          "accuracy",
          "Accuracy of normalized alignment parameters\n"
          "defined as 10^(-n), n>0.\n"
          "n = 6 is a good value."
        },

        /* Parameter 8 */
        { GIMP_PDB_INT32,
          "interpolation",
          "Interpolation alogrithm:\n"
          "0 ... none\n"
          "1 ... linear\n"
          "2 ... cubic\n"
          "3 ... lanczos"
        },

        /* Parameter 9 */
        { GIMP_PDB_INT32,
          "clip-results",
          "How to clip the results:\n"
          "0 ... adjust\n"
          "1 ... clip\n"
          "2 ... crop\n"
          "3 ... crop with aspect"
        }
    };

    /* Registration of textdomain for correct localization of menu
     * entries. */
    {
        gchar* locale_dir;
        
        locale_dir = Util_GetLocaleDir();
        if (locale_dir)
        {
            gimp_plugin_domain_register(PLUGIN_NAME, locale_dir);
            g_free(locale_dir);
        }
    }

    {
        gchar* data_dir;
        gchar* help_path;
        gchar* help_uri;
        data_dir = Util_GetDataDir();
        if (data_dir)
        {
            help_path = g_build_filename(data_dir, "help", NULL);
            if (help_path)
            {
                help_uri = g_filename_to_uri(help_path, NULL, NULL);
                if (help_uri)
                {
                    gimp_plugin_help_register(
                        "http://developer.gimp.org/" PLUGIN_NAME "/help",
                        help_uri);
                    g_free(help_uri);
                }
                g_free(help_path);
            }
            g_free(data_dir);
        }
    }
    
    gimp_install_procedure(
        /* name */
        PLUGIN_NAME,

        /* blurb */
        N_("Image Registration Tool: Transform (move,\n"
           "rotate, scale) the layer to minimize its\n"
           "difference to a given reference image"),

        /* help */
        N_("FIXME. Help message missing."),

        /* author */
        PLUGIN_AUTHOR,

        /* copyright */
        PLUGIN_COPYRIGHT,

        /* date */
        PLUGIN_DATE,
        
        /* menu label */
        N_("Image Registration..."),

        /* image types */
        "RGB*, GRAY*",

        /* procedure type */
        GIMP_PLUGIN,

        /* number of input parameters */
        G_N_ELEMENTS(args),
        
        /* number of output parameters */
        0,

        /* input parameters */
        args,
        
        /* output parameters */
        NULL);

    
    gimp_plugin_menu_register(
        PLUGIN_NAME,
        "<Image>/Tools");
}


void run(
    const gchar*     name,
    gint             nparams,
    const GimpParam* param,
    gint*            nreturn_vals,
    GimpParam**      return_vals)
{
    static GimpParam  values[1];
    GimpRunMode       run_mode;
    GimpPDBStatusType status;

    UserData user_data;
    gint32   image_id;
    gint32   drawable_id;

    GimpDrawable* drawable;
    
    if (!g_thread_supported())
    {
        g_thread_init(NULL);
    }

#if ENABLE_NLS
    /*  Initialize i18n support  */
    {
        gchar* locale_dir;
        locale_dir = Util_GetLocaleDir();
        if (locale_dir)
        {
            bindtextdomain (GETTEXT_PACKAGE, locale_dir);
            g_free(locale_dir);
            
# ifdef HAVE_BIND_TEXTDOMAIN_CODESET
            bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
# endif
            textdomain (GETTEXT_PACKAGE);
        }
    }
#endif

    run_mode    = param[0].data.d_int32;
    image_id    = param[1].data.d_int32;
    drawable_id = param[2].data.d_drawable;
    
    drawable    = gimp_drawable_get(drawable_id);

    /* Assume success. */
    status = GIMP_PDB_SUCCESS;
    
    if (run_mode == GIMP_RUN_NONINTERACTIVE)
    {
        if (nparams == 10)
        {
            user_data.ref_drawable   = param[3].data.d_drawable;
            user_data.matching_area  = param[4].data.d_int32;
            user_data.transformation = param[5].data.d_int32;
            user_data.num_iterations = param[6].data.d_int32;
            user_data.accuracy       = param[7].data.d_int32;
            user_data.interpolation  = param[8].data.d_int32;
            user_data.clip_method    = param[9].data.d_int32;
        }
        else
        {
            status = GIMP_PDB_CALLING_ERROR;
        }
    }
    else 
    {
        if (!gimp_procedural_db_get_data(PLUGIN_NAME, &user_data))
        {
            UserData_GetDefaults(&user_data);
        }

        if (run_mode == GIMP_RUN_INTERACTIVE)
        {
            gboolean success;
            success = RunDialog(drawable, &user_data);
            if (!success)
            {
                status = GIMP_PDB_CANCEL;
            }
        }
    }

    if (status == GIMP_PDB_SUCCESS)
    {
        gboolean success;
        success = Util_CheckParameters(drawable_id, &user_data);

        if (!success)
        {
            status = GIMP_PDB_CALLING_ERROR;
        }
        else
        {
            gdouble current_progress = 0.0;
            
            gimp_progress_init(
                _("Calculating image registration parameters..."));
            gimp_tile_cache_ntiles(
                (drawable->height + gimp_tile_height() - 1) /
                gimp_tile_height());
            success = Render(
                drawable,
                &user_data,
                ReportRenderProgress,
                &current_progress);
            gimp_progress_end();
            
            if (success)
            {
                if (run_mode != GIMP_RUN_NONINTERACTIVE)
                {
                    gimp_displays_flush();
                }
                if (run_mode == GIMP_RUN_INTERACTIVE)
                {
                    gimp_procedural_db_set_data(
                        PLUGIN_NAME,
                        &user_data,
                        sizeof(user_data));
                }
            }
            else
            {
                status = GIMP_PDB_EXECUTION_ERROR;
            }
        }
    }

    gimp_drawable_detach(drawable);
    
    /* Setting output values */
    *nreturn_vals = 1;
    *return_vals  = values + 0;

    Util_SetGimpStatus(values + 0, status);
}
