/*
 * gimp-image-reg -- A GIMP Plugin for Image Registration
 *
 * matrix.c
 * Copyright (C) 2010 Behnam Tabatabai
 *
 * This file is part of gimp-image-reg.
 *
 * gimp-image-reg is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * gimp-image-reg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with gimp-image-reg. If not, see <http://www.gnu.org/licenses/>.
 */

#include "matrix.h"

#include <string.h>

Matrix_Uint8* CreateMatrix_Uint8(
    gint32 num_rows,
    gint32 num_cols,
    gint32 num_chns)
{
    Matrix_Uint8* matrix;
    guint8* elements;
    
    if (num_rows < 1) return NULL;
    if (num_cols < 1) return NULL;
    if (num_chns < 1) return NULL;

    matrix = g_malloc(sizeof(Matrix_Uint8));
    if (!matrix) return NULL;
    
    elements = g_malloc0(num_cols * num_rows * num_chns * sizeof(guint8));

    if (elements)
    {
        matrix->num_rows = num_rows;
        matrix->num_cols = num_cols;
        matrix->num_chns = num_chns;
        matrix->elements = elements;
        return matrix;
    }
    else
    {
        g_free(matrix);
        return NULL;
    }
}

Matrix_Int16* CreateMatrix_Int16(
    gint32 num_rows,
    gint32 num_cols,
    gint32 num_chns)
{
    Matrix_Int16* matrix;
    gint16* elements;
    
    if (num_rows < 1) return NULL;
    if (num_cols < 1) return NULL;
    if (num_chns < 1) return NULL;

    matrix = g_malloc(sizeof(Matrix_Int16));
    if (!matrix) return NULL;
    
    elements = g_malloc0(num_cols * num_rows * num_chns * sizeof(gint16));

    if (elements)
    {
        matrix->num_rows = num_rows;
        matrix->num_cols = num_cols;
        matrix->num_chns = num_chns;
        matrix->elements = elements;
        return matrix;
    }
    else
    {
        g_free(matrix);
        return NULL;
    }
}

Matrix_Real* CreateMatrix_Real(
    gint32 num_rows,
    gint32 num_cols,
    gint32 num_chns)
{
    Matrix_Real* matrix;
    gdouble* elements;
    
    if (num_rows < 1) return NULL;
    if (num_cols < 1) return NULL;
    if (num_chns < 1) return NULL;

    matrix = g_malloc(sizeof(Matrix_Real));
    if (!matrix) return NULL;
    
    elements = g_malloc0(num_cols * num_rows * num_chns * sizeof(gdouble));

    if (elements)
    {
        matrix->num_rows = num_rows;
        matrix->num_cols = num_cols;
        matrix->num_chns = num_chns;
        matrix->elements = elements;
        return matrix;
    }
    else
    {
        g_free(matrix);
        return NULL;
    }
}

void DeleteMatrix_Uint8(Matrix_Uint8* matrix)
{
    if (matrix)
    {
        g_free(matrix->elements);
        g_free(matrix);
    }
}

void DeleteMatrix_Int16(Matrix_Int16* matrix)
{
    if (matrix)
    {
        g_free(matrix->elements);
        g_free(matrix);
    }
}

void DeleteMatrix_Real(Matrix_Real* matrix)
{
    if (matrix)
    {
        g_free(matrix->elements);
        g_free(matrix);
    }
}

void MatrixSetIdentity(Matrix_Real* matrix)
{
    gint i, m;

    g_assert(matrix && matrix->elements);
    g_assert(matrix->num_cols == matrix->num_rows);
    
    MatrixSetZero(matrix);

    for (i = 0; i < matrix->num_cols; ++i)
    {
        for (m = 0; m < matrix->num_chns; ++m)
        {
            *ELEM(matrix, i, i, m) = 1.0;
        }
    }
}

void MatrixSetZero(Matrix_Real* matrix)
{
    g_assert(matrix && matrix->elements);
    memset(
        matrix->elements,
        0,
        matrix->num_rows
        * matrix->num_cols
        * matrix->num_chns
        * sizeof(gdouble));
}

void MatrixMultiply(
    const Matrix_Real* a,
    const Matrix_Real* b,
    Matrix_Real* c)
{
    gint i, j, k, m;
    
    g_assert(a && a->elements);
    g_assert(b && b->elements);
    g_assert(c && b->elements);

    g_assert(NUM_COLS(a) == NUM_ROWS(b));
    g_assert(NUM_ROWS(a) == NUM_ROWS(c));
    g_assert(NUM_COLS(b) == NUM_COLS(c));
    g_assert(NUM_CHNS(a) == NUM_CHNS(b));
    g_assert(NUM_CHNS(a) == NUM_CHNS(c));
    
    for (i = 0; i < NUM_ROWS(a); ++i)
    {
        for (j = 0; j < NUM_COLS(b); ++j)
        {
            for (m = 0; m < NUM_CHNS(a); ++m)
            {
                *ELEM(c, i, j, m) = 0.0;
                for (k = 0; k < NUM_COLS(a); ++k)
                {
                    *ELEM(c, i, j, m) +=
                        *ELEM(a, i, k, m) *
                        *ELEM(a, k, j, m);
                }
            }
        }
    }
}
