/*
 * gimp-image-reg -- A GIMP Plugin for Image Registration
 *
 * render.c
 * Copyright (C) 2010 Behnam Tabatabai
 *
 * This file is part of gimp-image-reg.
 *
 * gimp-image-reg is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * gimp-image-reg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with gimp-image-reg. If not, see <http://www.gnu.org/licenses/>.
 */

#include "render.h"

#include "i18n.h"
#include "matrix.h"
#include "progress.h"
#include "warp.h"

/* --------------------------------------------------------------------
 * Render.
 * -------------------------------------------------------------------- */
gboolean Render(
    GimpDrawable*    drawable,
    const UserData*  user_data,
    ProgressCallback progress_callback,
    void*            client_data)
{
    Matrix_Real* warp;
    gboolean success;
    warp = CreateMatrix_Real(3, 3, 1);
    if (!warp)
    {
        return FALSE;
    }
    MatrixSetIdentity(warp);

    
    success = Warp(
        drawable,
        user_data,
        warp,
        progress_callback,
        client_data);
    
    /* ------------------------------------------------------------------
     *  Actual rendering begins here.
     * ------------------------------------------------------------------ */
    if (success)
    {
       gint32 image_id;
       gint32 active_layer;
       gint32 saved_selection;
       
       image_id = gimp_drawable_get_image(drawable->drawable_id);
       active_layer = gimp_image_get_active_layer(image_id);
       
       success = gimp_image_undo_group_start(image_id);
       if (!success)
       {
           DeleteMatrix_Real(warp);
           return FALSE;
       }

       /* Save the selection. */
       saved_selection = -1;
       if (!gimp_selection_is_empty(image_id))
       {
           saved_selection = gimp_selection_save(image_id);
           gimp_selection_none(image_id);
       }

       if (progress_callback)
       {
           gboolean cont;
           cont = progress_callback(
               client_data,
               _("transforming image"),
               0, 0.0, 0.0);
           if (!cont)
           {
               return FALSE;
           }
       }
       
       gimp_drawable_transform_matrix(
           drawable->drawable_id,
           *ELEM(warp, 0, 0, 0),
           *ELEM(warp, 0, 1, 0),
           *ELEM(warp, 0, 2, 0),
           *ELEM(warp, 1, 0, 0),
           *ELEM(warp, 1, 1, 0),
           *ELEM(warp, 1, 2, 0),
           *ELEM(warp, 2, 0, 0),
           *ELEM(warp, 2, 1, 0),
           *ELEM(warp, 2, 2, 0),
           GIMP_TRANSFORM_FORWARD,
           user_data->interpolation,
           FALSE, /* super-sampling, will be ignored */
           3,     /* recursion level used in super-sampling */
           user_data->clip_method);

       if (gimp_drawable_is_valid(saved_selection))
       {
           gimp_selection_load(saved_selection);
       }
       gimp_image_set_active_layer(image_id, active_layer);
       
       /* We do not care what these functions return. */
       gimp_image_undo_group_end(image_id);
       gimp_displays_flush();
    }
    
    /* Release the locally allocated resources. */
    DeleteMatrix_Real(warp);

    return success;
}
