/*
 * gimp-image-reg -- A GIMP Plugin for Image Registration
 *
 * warp.c
 * Copyright (C) 2010 Behnam Tabatabai
 *
 * This file is part of gimp-image-reg.
 *
 * gimp-image-reg is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * gimp-image-reg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with gimp-image-reg. If not, see <http://www.gnu.org/licenses/>.
 */

#include "warp.h"

#include "i18n.h"
#include "image-ops.h"
#include "linalg.h"

#include <math.h>
#include <stdio.h>
#include <string.h>


#ifndef NUM_THREADS
# define NUM_THREADS 4
#endif

#ifndef MULTIPLE_THREADS
# define MULTIPLE_THREADS 1
#endif

/* --------------------------------------------------------------------
 * Data types.
 * -------------------------------------------------------------------- */

/* The data passed to ComputeRhs. */
typedef struct {
    /* Input. */
    const Image* the_image;
    const Image* ref_image;
    const Matrix_Real* warp;
    gint32 wmin; /* inclusive */
    gint32 wmax; /* exclusive */
    gint32 hmin; /* inclusive */
    gint32 hmax; /* exclusive */
    gint dof;
    /* Output. */
    gdouble rhs[6];
    gint32 num_pixels;
    gint32 num_pixels_inside;
} ComputationData;

#if MULTIPLE_THREADS
/* The data passed to ThreadFunction. */
typedef struct {
    GThread*         thread;
    GCond*           cond_1;
    GCond*           cond_2;
    GMutex*          mutex_1;
    GMutex*          mutex_2;
    gint             code_1;
    gint             code_2;
    ComputationData* comp_data;
} ThreadData;
#endif

/* --------------------------------------------------------------------
 * Prototype of the local functions.
 * -------------------------------------------------------------------- */
#if MULTIPLE_THREADS
static void PrepareThreads(
    ComputationData* comp_data,
    ThreadData* thread_data,
    gint num_threads);

static void ClearSyncObjs(ThreadData* thread_data, gint num_threads);

static void SignalThreads(
    ThreadData* thread_data,
    gint        num_threads,
    gint        code);

static void WaitForCompletion(ThreadData* thread_data, gint num_threads);
static void WaitForExit(ThreadData* thread_data, gint num_threads);

static gpointer ThreadFunction(gpointer data);
#endif

static gdouble PercentualAccuracy(
    gdouble err,
    gint32  accuracy);

static void ComputeRhs(ComputationData* data);

static gdouble NormalizationFactor(
    gint32 width,
    gint32 height);

static void MakeWarpMatrix(
    Matrix_Real* warp,
    const gdouble* params,
    gint dof,
    gint32 width,
    gint32 height);

static gboolean Interpolate(
    const Image* image,
    const gdouble* coord,
    gdouble* value);

static void TransformPoint(
    const Matrix_Real* warp,
    const gdouble*     x,
    gdouble*           z);


/* --------------------------------------------------------------------
 * Local functions.
 * -------------------------------------------------------------------- */
#if MULTIPLE_THREADS
void PrepareThreads(
    ComputationData* comp_data,
    ThreadData* thread_data,
    gint num_threads)
{
    gint i;
    ThreadData* t;
    ComputationData* c;
    for (c = comp_data, t = thread_data, i = 0; i < num_threads; ++c, ++t, ++i)
    {
        t->comp_data = c;
        t->cond_1    = g_cond_new();
        t->cond_2    = g_cond_new();
        t->mutex_1   = g_mutex_new();
        t->mutex_2   = g_mutex_new();
        t->code_1    = 0;
        t->code_2    = 0;

        t->thread    = g_thread_create(
            ThreadFunction, /* thread function */
            t,              /* pointer passed to thread function */
            TRUE,           /* joinable */
            NULL);
    }
}

void ClearSyncObjs(ThreadData* thread_data, gint num_threads)
{
    gint i;
    ThreadData* t;
    for (t = thread_data, i = 0; i < num_threads; ++t, ++i)
    {
        g_mutex_free(t->mutex_1);
        g_mutex_free(t->mutex_2);
        g_cond_free(t->cond_1);
        g_cond_free(t->cond_2);
    }
}

void SignalThreads(
    ThreadData* thread_data,
    gint        num_threads,
    gint        code)
{
    gint i;
    ThreadData* t;
    for (t = thread_data, i = 0; i < num_threads; ++t, ++i)
    {
        g_mutex_lock(t->mutex_1);
        t->code_1 = code;
        g_cond_signal(t->cond_1);
        g_mutex_unlock(t->mutex_1);
    }
}


void WaitForCompletion(ThreadData* thread_data, gint num_threads)
{
    gint i;
    ThreadData* t;
    for (t = thread_data, i = 0; i < num_threads; ++t, ++i)
    {
        g_mutex_lock(t->mutex_2);
        while (t->code_2 == 0)
        {
            g_cond_wait(t->cond_2, t->mutex_2);
        }
        if (t->code_2 == 1)
        {
            // Ok.
        }
        else
        {
            g_assert(!"Program error.");
        }
        t->code_2 = 0;
        g_mutex_unlock(t->mutex_2);
    }
}

void WaitForExit(ThreadData* thread_data, gint num_threads)
{
    gint i;
    ThreadData* t;
    for (t = thread_data, i = 0; i < num_threads; ++t, ++i)
    {
        g_thread_join(t->thread);
    }
}

gpointer ThreadFunction(gpointer thread_data)
{
    ThreadData* data = (ThreadData*) thread_data;
    g_assert(data);
    while (1)
    {
        g_mutex_lock(data->mutex_1);
        while (data->code_1 == 0)
        {
            g_cond_wait(data->cond_1, data->mutex_1);
        }

        if (data->code_1 == 1)
        {
            ComputeRhs(data->comp_data);

            g_mutex_lock(data->mutex_2);
            data->code_2 = 1;
            g_cond_signal(data->cond_2);
            g_mutex_unlock(data->mutex_2);

            data->code_1 = 0;
            g_mutex_unlock(data->mutex_1);
        }
        else if (data->code_1 == 2)
        {
            data->code_1 = 0;
            g_mutex_unlock(data->mutex_1);
            break;
        }
        else
        {
            g_assert(!"Program error.");
        }
    }
    return NULL;
}
#endif

void ComputeRhs(ComputationData* data)
{
    gint32 r, c;
    gint m;

    const guint8*  pixel;
    const gint16*  gradient;
    const guint8*  mask = NULL;
    const gdouble* steepest;

    gdouble coord[3];
    gdouble ref_coord[3];

    gboolean inside;
    gdouble ref_value;
    gdouble diff;

    g_assert(data);

    data->num_pixels = 0;
    data->num_pixels_inside = 0;

    for (m = 0; m < data->dof; ++m)
    {
        data->rhs[m] = 0.0;
    }

    coord[2] = 1.0;

    for (c = data->wmin; c < data->wmax; c += 1)
    {
        pixel    = ELEM(data->the_image->pixel,            data->hmin, c, 0);
        gradient = ELEM(data->the_image->gradient,         data->hmin, c, 0);
        steepest = ELEM(data->the_image->steepest_descent, data->hmin, c, 0);
        if (data->the_image->mask)
        {
            mask = ELEM(data->the_image->mask, data->hmin, c, 0);
        }

        /* Adjust the offsets, so that both images have the same
         * coordinate frames. See also the adjustment of the
         * y-coordinate below. */

        coord[0] = (gdouble)(c
                             + data->the_image->x_roi
                             + data->the_image->x_offset
                             - data->ref_image->x_offset);

        for (r = data->hmin; r < data->hmax;
                 r        += 1,
                 pixel    += 1,
                 gradient += 2,
                 steepest += data->dof)
        {

            if (!mask || *mask)
            {
                data->num_pixels += 1;

                coord[1] = (gdouble)(r
                                     + data->the_image->y_roi
                                     + data->the_image->y_offset
                                     - data->ref_image->y_offset);

                TransformPoint(data->warp, coord, ref_coord);

                inside = Interpolate(
                    data->ref_image,
                    ref_coord,
                    &ref_value);

                if (inside)
                {
                    data->num_pixels_inside += 1;
                    diff = ref_value - (gdouble)(*pixel);
                    for (m = 0; m < data->dof; ++m)
                    {
                        data->rhs[m] += steepest[m] * diff;
                    }
                }

            }
            if (mask)
            {
                mask += 1;
            }
        }
    }
}

gdouble PercentualAccuracy(
    gdouble err,
    gint32  accuracy)
{

    gdouble err_0;

    gdouble log_err;
    gdouble log_err_0;

    err_0 = 1.0 / pow(10.0, accuracy);

    if (err < err_0)
    {
        return 1.0;
    }

    if (err > 1.0)
    {
        return 0.0;
    }

    log_err   = log(err);
    log_err_0 = log(err_0);

    return log_err / log_err_0;

}


gdouble NormalizationFactor(
    gint32 width,
    gint32 height)
{
    if (width > height)
    {
        return width;
    }
    else
    {
        return height;
    }
}


void MakeWarpMatrix(
    Matrix_Real* warp,
    const gdouble* params,
    gint dof,
    gint32 width,
    gint32 height)
{
    gdouble norm_factor;

    g_assert(warp && warp->elements);
    g_assert(NUM_ROWS(warp) == 3);
    g_assert(NUM_COLS(warp) == 3);
    g_assert(NUM_CHNS(warp) == 1);
    g_assert(params);
    g_assert(dof == 3 || dof == 6);

    norm_factor = NormalizationFactor(width, height);

    MatrixSetIdentity(warp);

    if (dof == 3)
    {
        gdouble p;
        p = params[0] * norm_factor;
        if (fabs(p) > 1.0)
        {
            p = (p > 1.0) ? 1.0 : -1.0;
        }
        *ELEM(warp, 0, 0, 0) = sqrt(1.0 - p*p);
        *ELEM(warp, 1, 1, 0) = sqrt(1.0 - p*p);
        *ELEM(warp, 0, 1, 0) = -p;
        *ELEM(warp, 1, 0, 0) =  p;
        *ELEM(warp, 0, 2, 0) =   params[1];
        *ELEM(warp, 1, 2, 0) =   params[2];
    }
    else
    {
        *ELEM(warp, 0, 0, 0) += params[0] * norm_factor;
        *ELEM(warp, 1, 0, 0)  = params[1] * norm_factor;
        *ELEM(warp, 0, 1, 0)  = params[2] * norm_factor;
        *ELEM(warp, 1, 1, 0) += params[3] * norm_factor;
        *ELEM(warp, 0, 2, 0)  = params[4];
        *ELEM(warp, 1, 2, 0)  = params[5];
    }
}

gboolean Interpolate(
    const Image* image,
    const gdouble* coord,
    gdouble* value)
{
    gdouble u, v;
    gdouble du, dv;
    gdouble n[4];
    gdouble f[4];
    gint32 x, y;
    const guint8* pixel;

    u = coord[0] - image->x_roi;
    v = coord[1] - image->y_roi;

    x = (gint32)(floor(u) + 0.5);
    y = (gint32)(floor(v) + 0.5);

    if (x < 0) return FALSE;
    if (y < 0) return FALSE;
    if (x >= image->w_roi-1) return FALSE;
    if (y >= image->h_roi-1) return FALSE;


    du = (u - x - 0.5) * 2.0;
    dv = (v - y - 0.5) * 2.0;

    n[0] = (1.0 - du) * (1.0 - dv);
    n[1] = (1.0 - du) * (1.0 + dv);
    n[2] = (1.0 + du) * (1.0 + dv);
    n[3] = (1.0 + du) * (1.0 - dv);

    pixel = ELEM(image->pixel, y, x, 0);

    f[0] = *(pixel + 0);
    f[1] = *(pixel + 1);
    f[2] = *(pixel + 1 + image->h_roi);
    f[3] = *(pixel + image->h_roi);

    *value = (
        n[0] * f[0] +
        n[1] * f[1] +
        n[2] * f[2] +
        n[3] * f[3] ) * 0.25;

    return TRUE;
}

void TransformPoint(
    const Matrix_Real* warp,
    const gdouble*     x,
    gdouble*           z)
{
    gint i, j;
    const gdouble* w;
    g_assert(warp && warp->elements);
    g_assert(NUM_ROWS(warp) == 3);
    g_assert(NUM_COLS(warp) == 3);
    g_assert(NUM_CHNS(warp) == 1);

    for (i = 0; i < 2; ++i)
    {
        z[i] = 0.0;
        w = ELEM(warp, i, 0, 0);
        for (j = 0; j < 3; ++j, w+=3)
        {
            z[i] += *w * x[j];
        }
    }
    z[2] = 1.0;
}




gboolean Warp(
    GimpDrawable* drawable,
    const UserData* user_data,
    Matrix_Real* warp,
    ProgressCallback progress_callback,
    void* client_data)
{
    gboolean success;
    Image the_image;
    Image ref_image;

    Matrix_Real* delta_warp;

    /* Degrees of freedom. Will be 3 or 6 */
    gint dof;

    /* Large enough for 3 or 6 degrees of freedom. */
    gint permutation_vector[6];
    gdouble rhs[6];
    gdouble delta_p[6];

    gint iter;
    gdouble err;

    ComputationData comp_data[NUM_THREADS];

#if MULTIPLE_THREADS
    ThreadData thread_data[NUM_THREADS];
#endif

    if (progress_callback)
    {
        gboolean cont;
        cont = progress_callback(
            client_data,
            _("init"),
            0, 0.0, 0.0);
        if (!cont)
        {
            return FALSE;
        }
    }

    /* ------------------------------------------------------------------
     *  Prepare the data for the image.
     * ------------------------------------------------------------------ */

    /* (0) Set everything to zero. */
    memset(&the_image, 0, sizeof(Image));

    /* (1) Width and height. */
    the_image.width  = drawable->width;
    the_image.height = drawable->height;

    /* (2) The offsets. */
    success = gimp_drawable_offsets(
        drawable->drawable_id,
        &the_image.x_offset,
        &the_image.y_offset);

    if (!success)
    {
        return FALSE;
    }

    /* (3) Region of interest. */
    switch (user_data->matching_area)
    {
        case MATCHING_AREA_WHOLE_IMAGE:
        {
            the_image.x_roi = 0;
            the_image.y_roi = 0;
            the_image.w_roi = drawable->width;
            the_image.h_roi = drawable->height;
        } break;

        case MATCHING_AREA_SELECTION:
        {
            gboolean non_empty;
            non_empty = gimp_drawable_mask_intersect(
                drawable->drawable_id,
                &the_image.x_roi,
                &the_image.y_roi,
                &the_image.w_roi,
                &the_image.h_roi);

            if (!non_empty)
            {
                /* It should be impossible to select this option if
                 * there is a selection and its intersection with the
                 * drawable is empty (see dialog.c). Anyway, it looks
                 * like a program error. */
                return FALSE;
            }
        } break;

        default:
        {
            /* We should never get here. */
            return FALSE;
        }
    }

    /* (4) Pixel data. */
    if (progress_callback)
    {
        gboolean cont;
        cont = progress_callback(
            client_data,
            _("pixel data"),
            0, 0.0, 0.0);
        if (!cont)
        {
            return FALSE;
        }
    }
    the_image.pixel = Extract_GrayScaleImage(
        drawable,
        the_image.x_roi,
        the_image.y_roi,
        the_image.w_roi,
        the_image.h_roi);

    if (!the_image.pixel)
    {
        return FALSE;
    }

    /* (5) Selection mask. */
    if (progress_callback)
    {
        gboolean cont;
        cont = progress_callback(
            client_data,
            _("selection mask"),
            0, 0.0, 0.0);
        if (!cont)
        {
            return FALSE;
        }
    }
    switch (user_data->matching_area)
    {
        case MATCHING_AREA_WHOLE_IMAGE:
        {
            /* Nothing to do for this case. */
        } break;

        case MATCHING_AREA_SELECTION:
        {
            gint32 image_id;
            gint32 channel_id;
            GimpDrawable* channel;

            image_id = gimp_drawable_get_image(drawable->drawable_id);
            channel_id = gimp_image_get_selection(image_id);
            if (!gimp_drawable_is_channel(channel_id))
            {
                ReleaseImage(&the_image);
                return FALSE;
            }
            channel = gimp_drawable_get(channel_id);
            if (!channel)
            {
                ReleaseImage(&the_image);
                return FALSE;
            }

            the_image.mask = Extract_GrayScaleImage(
                channel,
                the_image.x_roi + the_image.x_offset,
                the_image.y_roi + the_image.y_offset,
                the_image.w_roi,
                the_image.h_roi);

            gimp_drawable_detach(channel);

            if (!the_image.mask)
            {
                ReleaseImage(&the_image);
                return FALSE;
            }
        } break;

        default:
        {
            /* We should never get here. */
            ReleaseImage(&the_image);
            return FALSE;
        }
    }

    /* (6) Gradient. */
    if (progress_callback)
    {
        gboolean cont;
        cont = progress_callback(
            client_data,
            _("gradient"),
            0, 0.0, 0.0);
        if (!cont)
        {
            return FALSE;
        }
    }
    the_image.gradient = CreateGradient(the_image.pixel);
    if (!the_image.gradient)
    {
        ReleaseImage(&the_image);
        return FALSE;
    }

    /* (7) Hessian and steepest descent components. */
    if (progress_callback)
    {
        gboolean cont;
        cont = progress_callback(
            client_data,
            _("hessian"),
            0, 0.0, 0.0);
        if (!cont)
        {
            return FALSE;
        }
    }
    {
        gint m1, m2;
        gint32 r, c;
        gint32 x, y;
        const gint16* gradient;
        const guint8* mask = NULL;
        gdouble gx, gy;
        gdouble* steepest;

        gdouble norm_factor;
        norm_factor = NormalizationFactor(the_image.width, the_image.height);

        switch (user_data->transformation)
        {
            case XFORMATION_RIGID:
            {
                dof = 3;
            } break;
            case XFORMATION_AFFINE:
            {
                dof = 6;
            } break;
            default:
            {
                /* We should never get here. */
                ReleaseImage(&the_image);
                return FALSE;
            }
        }

        the_image.steepest_descent = CreateMatrix_Real(
            the_image.h_roi, /* number of rows */
            the_image.w_roi, /* number of cols */
            dof              /* number of chns */ );

        if (!the_image.steepest_descent)
        {
            ReleaseImage(&the_image);
            return FALSE;
        }

        the_image.hessian = CreateMatrix_Real(dof, dof, 1);
        if (!the_image.hessian)
        {
            ReleaseImage(&the_image);
            return FALSE;
        }

        for (c = 0; c < the_image.w_roi; ++c)
        {
            x = c + the_image.x_roi;

            gradient = ELEM(the_image.gradient, 0, c, 0);
            steepest = ELEM(the_image.steepest_descent, 0, c, 0);
            if (the_image.mask)
            {
                mask = ELEM(the_image.mask, 0, c, 0);
            }

            if (dof == 3)
            {
                for (r = 0;
                     r < the_image.h_roi;
                     ++r, gradient += 2, steepest += dof)
                {
                    if (!mask || *mask)
                    {
                        y = r + the_image.y_roi;

                        gx = *(gradient + 0);
                        gy = *(gradient + 1);

                        *(steepest + 0) = (x * gy - y * gx) * norm_factor;
                        *(steepest + 1) = gx;
                        *(steepest + 2) = gy;

                        for (m1 = 0; m1 < 3; ++m1)
                        {
                            for (m2 = 0; m2 < 3; ++m2)
                            {
                                *ELEM(the_image.hessian, m1, m2, 0) +=
                                    *(steepest + m1) *
                                    *(steepest + m2);
                            }
                        }
                    }
                    if (mask)
                    {
                        mask += 1;
                    }
                }
            }
            else
            {
                for (r = 0;
                     r < the_image.h_roi;
                     ++r, gradient += 2, steepest += dof)
                {
                    if (!mask || *mask)
                    {
                        y = r + the_image.y_roi;

                        gx = *(gradient + 0);
                        gy = *(gradient + 1);

                        *(steepest + 0) = x * gx * norm_factor;
                        *(steepest + 1) = x * gy * norm_factor;
                        *(steepest + 2) = y * gx * norm_factor;
                        *(steepest + 3) = y * gy * norm_factor;
                        *(steepest + 4) = gx;
                        *(steepest + 5) = gy;

                        for (m1 = 0; m1 < 6; ++m1)
                        {
                            for (m2 = 0; m2 < 6; ++m2)
                            {
                                *ELEM(the_image.hessian, m1, m2, 0) +=
                                    *(steepest + m1) *
                                    *(steepest + m2);
                            }
                        }
                    }
                    if (mask)
                    {
                        mask += 1;
                    }
                }
            }
        }
    }

    /* ------------------------------------------------------------------
     *  Prepare the data for the reference image.
     * ------------------------------------------------------------------ */
    {
        GimpDrawable* ref_drawable;

        if (progress_callback)
        {
            gboolean cont;
            cont = progress_callback(
                client_data,
                _("reference image"),
                0, 0.0, 0.0);
             if (!cont)
            {
                return FALSE;
            }
        }

        ref_drawable = gimp_drawable_get(user_data->ref_drawable);
        if (!ref_drawable)
        {
            ReleaseImage(&the_image);
            return FALSE;
        }

        /* (0) Set everything to zero. */
        memset(&ref_image, 0, sizeof(Image));

        /* (1) Width and height. */
        ref_image.width  = ref_drawable->width;
        ref_image.height = ref_drawable->height;

        /* (2) The offsets. */
        success = gimp_drawable_offsets(
            ref_drawable->drawable_id,
            &ref_image.x_offset,
            &ref_image.y_offset);

        if (!success)
        {
            ReleaseImage(&the_image);
            return FALSE;
        }

        /* (3) Region of interest. */
        ref_image.x_roi = 0;
        ref_image.y_roi = 0;
        ref_image.w_roi = ref_drawable->width;
        ref_image.h_roi = ref_drawable->height;

        /* (4) Pixel data. */
        ref_image.pixel = Extract_GrayScaleImage(
            ref_drawable,
            ref_image.x_roi,
            ref_image.y_roi,
            ref_image.w_roi,
            ref_image.h_roi);

        if (!ref_image.pixel)
        {
            return FALSE;
        }

        gimp_drawable_detach(ref_drawable);
    }

    /* ------------------------------------------------------------------
     *  Calculation of the transformation parameters.
     * ------------------------------------------------------------------ */
    if (progress_callback)
    {
        gboolean cont;
        cont = progress_callback(
            client_data,
            _("warp parameters"),
            0, 0.0, 0.0);
        if (!cont)
        {
            return FALSE;
        }
    }

    /* Inverse of the Hessian. The result is stored on the matrix
     * itself. */
    success = LU_Decompose(
        the_image.hessian,
        permutation_vector,
        NULL);

    if (!success)
    {
        ReleaseImage(&the_image);
        ReleaseImage(&ref_image);

        return FALSE;
    }


    delta_warp = CreateMatrix_Real(3, 3, 1);
    if (!delta_warp)
    {
        ReleaseImage(&the_image);
        ReleaseImage(&ref_image);

        return FALSE;
    }

    /* --------------------------------------------------------------------
     * Prepare the data for parallel computation.
     * -------------------------------------------------------------------- */
    {
        gint i;
        gint32 delta;
        if (the_image.w_roi < the_image.h_roi)
        {
            /* Divide along height. */
            delta = the_image.h_roi / NUM_THREADS;
        }
        else
        {
            /* Divide along width. */
            delta = the_image.w_roi / NUM_THREADS;
        }

        for (i = 0; i < NUM_THREADS; ++i)
        {
            comp_data[i].the_image = &the_image;
            comp_data[i].ref_image = &ref_image;
            comp_data[i].warp      = warp;
            comp_data[i].dof       = dof;
            if (the_image.w_roi < the_image.h_roi)
            {
                comp_data[i].wmin = 0;
                comp_data[i].wmax = the_image.w_roi;
                comp_data[i].hmin = i * delta;
                comp_data[i].hmax = i * delta + delta;
            }
            else
            {
                comp_data[i].wmin = i * delta;
                comp_data[i].wmax = i * delta + delta;
                comp_data[i].hmin = 0;
                comp_data[i].hmax = the_image.h_roi;
            }
        }
        if (the_image.w_roi < the_image.h_roi)
        {
            comp_data[NUM_THREADS-1].hmax = the_image.h_roi;
        }
        else
        {
            comp_data[NUM_THREADS-1].wmax = the_image.w_roi;
        }
    }

#if MULTIPLE_THREADS
    PrepareThreads(comp_data, thread_data, NUM_THREADS);
#endif


    iter = 0;

    while (TRUE)
    {
        gint t;
        gint m;

        gint num_pixels_roi;
        gint count_inside;

        /* Start the threads. */
#if MULTIPLE_THREADS
        SignalThreads(thread_data, NUM_THREADS, 1);
#else
        for (t = 0; t < NUM_THREADS; ++t)
        {
            ComputeRhs(comp_data + t);
        }
#endif

        for (m = 0; m < dof; ++m)
        {
            rhs[m] = 0.0;
        }

        /* Wait for completion and collect the results. */
#if MULTIPLE_THREADS
        WaitForCompletion(thread_data, NUM_THREADS);
#endif

        num_pixels_roi = 0;
        count_inside   = 0;

        for (t = 0; t < NUM_THREADS; ++t)
        {
            num_pixels_roi += comp_data[t].num_pixels;
            count_inside   += comp_data[t].num_pixels_inside;
            for (m = 0; m < dof; ++m)
            {
                rhs[m] += comp_data[t].rhs[m];
            }
        }

        if (count_inside > 0)
        {
            for (m = 0; m < dof; ++m)
            {
                rhs[m] *= (gdouble)(num_pixels_roi) / (gdouble)(count_inside);
            }
        }
        else
        {
            break;
        }

        /* Compositional parameter increment. */
        LU_BackSubstitute(
            the_image.hessian,
            permutation_vector,
            rhs, delta_p);

        MakeWarpMatrix(
            delta_warp,
            delta_p,
            dof,
            the_image.width,
            the_image.height);

        /* Calculation of warp = delta_warp^(-1) * warp */
        {
            gint perm[3];
            gdouble temp[3];
            gdouble* w;
            gint j;
            success = LU_Decompose(
                delta_warp,
                perm, NULL);

            if (!success)
            {
                DeleteMatrix_Real(warp);
                DeleteMatrix_Real(delta_warp);
                ReleaseImage(&the_image);
                ReleaseImage(&ref_image);

                return FALSE;
            }

            w = ELEM(warp, 0, 0, 0);
            for (j = 0; j < 3; ++j, w += 3)
            {
                LU_BackSubstitute(
                    delta_warp,
                    perm,
                    w, temp);
                memcpy(w, temp, sizeof(temp));
            }
        }

        err = 0.0;
        for (m = 0; m < dof; ++m)
        {
            err = MAX(err, fabs(delta_p[m]));
        }

        if (err < 1.0 / pow(10.0, user_data->accuracy))
        {
            break;
        }

        ++iter;
        if (progress_callback)
        {
            gboolean cont;
            cont = progress_callback(
                client_data,
                NULL,
                iter, err,
                PercentualAccuracy(err, user_data->accuracy));
            if (!cont)
            {
                return FALSE;
            }
        }

        if (iter == user_data->num_iterations)
        {
            break;
        }
    }

#if MULTIPLE_THREADS
    SignalThreads(thread_data, NUM_THREADS, 2);
    WaitForExit(thread_data, NUM_THREADS);
    ClearSyncObjs(thread_data, NUM_THREADS);
#endif

    /* Release the locally allocated resources. */
    DeleteMatrix_Real(delta_warp);
    ReleaseImage(&the_image);
    ReleaseImage(&ref_image);

    return TRUE;
}

