/* GIMP Plug-in Template
 * Copyright (C) 2000  Michael Natterer <mitch@gimp.org> (the "Author").
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of the Author of the
 * Software shall not be used in advertising or otherwise to promote the
 * sale, use or other dealings in this Software without prior written
 * authorization from the Author.
 */

#include "config.h"

#include <gtk/gtk.h>

#include <libgimp/gimp.h>

#include "main.h"
#include "render.h"

#include "plugin-intl.h"


/*  Public functions  */

void
render (gint32              image_ID,
	GimpDrawable       *drawable,
	PlugInVals         *vals,
	PlugInImageVals    *image_vals,
	PlugInDrawableVals *drawable_vals)
{
  GimpPixelRgn src_rgn, dst_rgn, msk_rgn;
  gint    progress, max_progress;
  gint    x1, y1, x2, y2;
  gint    x, y, w, h;
  gint    rgn_stride, msk_stride;
  gint    rgn_bpp, msk_bpp;
  guchar  *rgn_buf, *msk_buf;
  guchar  **src, **dst, **msk;

  gboolean do_hor = vals->do_hor;
  gboolean do_ver = vals->do_ver;

  gint32 selection_ID;
  GimpDrawable* msk_drawable;

  /* Get selection */
  selection_ID = gimp_image_get_selection (image_ID);

  msk_drawable = gimp_drawable_get (selection_ID);
  msk_bpp   = gimp_drawable_bpp (msk_drawable->drawable_id);

  /* Get selection area */
  gimp_drawable_mask_bounds (drawable->drawable_id, &x1, &y1, &x2, &y2);

  rgn_bpp   = gimp_drawable_bpp (drawable->drawable_id);

  w = (x2 - x1);
  h = (y2 - y1);

  /* Initialize progress */
  progress = 0;
  max_progress = w * h;
  if (do_hor && do_ver) max_progress *= 2;

  /* substitute pixel vales */

  gimp_pixel_rgn_init (&src_rgn, drawable,
                       x1, y1, w, h, FALSE, FALSE);
  gimp_pixel_rgn_init (&dst_rgn, drawable,
                       x1, y1, w, h, TRUE, TRUE);
  gimp_pixel_rgn_init (&msk_rgn, msk_drawable,
		       x1, y1, w, h, FALSE, FALSE);

  rgn_buf = g_new0 (guchar, w * h * rgn_bpp);
  msk_buf = g_new (guchar, w * h * msk_bpp);

  gimp_progress_init (_("Intelligum"));

  gimp_pixel_rgn_get_rect(&src_rgn, rgn_buf, x1, y1, w, h);
  gimp_pixel_rgn_get_rect(&msk_rgn, msk_buf, x1, y1, w, h);

  rgn_stride = w * rgn_bpp;
  msk_stride = w * msk_bpp;

  src = g_new (guchar*, h);
  dst = g_new (guchar*, h);
  msk = g_new (guchar*, h);

  for(y = 0; y < h; y++) {
    src[y] = &rgn_buf[y * rgn_stride];
    dst[y] = &rgn_buf[y * rgn_stride];
    msk[y] = &msk_buf[y * msk_stride];
  }

  if (do_hor)
    for (y = 1; y < h-1; y++)
      {
	gint    xl, xr, xm;
	xl = xr = 0;
	x = 1;
	while (x<w-1)
	  {
	    while ((x<w)&&(msk[y][x]==0)) x++;
	    xl = x;
	    while ((x<w)&&(msk[y][x]!=0)) x++;
	    xr = x;
	    if (xr>=w) xr=w-1;
	    for(xm=xl+1; xm<xr-1; xm++) {
	      gdouble wm;
	      gint col;
	      wm = ((gdouble)(xr-xm)/(gdouble)(xr-xl));
	      for(col=0;col<rgn_bpp;col++)
		{
		  dst[y][xm*rgn_bpp+col]=(guchar)CLAMP0255((gdouble)src[y][(xl)*rgn_bpp+col]*(wm)+
							   (gdouble)src[y][(xr)*rgn_bpp+col]*(1.0-wm));
		}
	    }
	  }

	/* Update progress */
	progress += w;

	gimp_progress_update ((double) progress / (double) max_progress);
      }

  if (do_ver) {
    for (x = 1; x < w-1; x++)
      {
	gint    yt, yb, ym;
	yt = yb = 0;
	y = 1;
	while (y<h-1)
	  {
	    while ((y<h-1)&&(msk[y][x]==0)) y++;
	    yb = y;
	    while ((y<h-1)&&(msk[y][x]!=0)) y++;
	    yt = y;
	    if (yt>=h-1) yt=h-1;
	    for(ym=yb+1; ym<yt-1; ym++) {
	      gdouble wm;
	      gint col;
	      wm = ((gdouble)(yt-ym)/(gdouble)(yt-yb));
	      for(col=0;col<rgn_bpp;col++)
		{
		  if (!do_hor)
		    dst[ym][x*rgn_bpp+col]=(guchar)CLAMP0255(((gdouble)(src[yb][x*rgn_bpp+col]*(wm))+
							      (gdouble)(src[yt][x*rgn_bpp+col])*(1.0-wm)));
		  else
		    dst[ym][x*rgn_bpp+col]=(guchar)CLAMP0255(((gdouble)dst[ym][x*rgn_bpp+col])/2+
							     ((gdouble)(src[yb][x*rgn_bpp+col]*(wm))+
							      (gdouble)(src[yt][x*rgn_bpp+col])*(1.0-wm))/2);
		}
	    }
	  }

	/* Update progress */
	progress += w;

	gimp_progress_update ((double) progress / (double) max_progress);
      }
  }
  gimp_pixel_rgn_set_rect(&dst_rgn, rgn_buf, x1, y1, w, h);

  g_free(src);
  g_free(dst);
  g_free(msk);

  g_free(rgn_buf);
  g_free(msk_buf);

  gimp_progress_update (1.0);

  /*  update the region  */
  gimp_drawable_flush (drawable);
  gimp_drawable_merge_shadow (drawable->drawable_id, TRUE);
  gimp_drawable_update (drawable->drawable_id, x1, y1, w, h);
}
